# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_is_fitted, check_array

from ._tree_builder import ClassificationShapeletTreeBuilder
from ._tree_builder import ExtraClassificationShapeletTreeBuilder
from ._tree_builder import ExtraRegressionShapeletTreeBuilder
from ._tree_builder import RegressionShapeletTreeBuilder
from ..distance._distance import _DISTANCE_MEASURE


class BaseShapeletTree(BaseEstimator):
    def __init__(
        self,
        *,
        max_depth=None,
        min_samples_split=2,
        n_shapelets=10,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="euclidean",
        metric_params=None,
        force_dim=None,
        random_state=None
    ):
        if min_shapelet_size < 0 or min_shapelet_size > max_shapelet_size:
            raise ValueError(
                "`min_shapelet_size` {0} <= 0 or {0} > {1}".format(
                    min_shapelet_size, max_shapelet_size
                )
            )
        if max_shapelet_size > 1:
            raise ValueError("`max_shapelet_size` {0} > 1".format(max_shapelet_size))
        self.max_depth = max_depth or 2 ** 31
        self.min_samples_split = min_samples_split
        self.random_state = check_random_state(random_state)
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metric = metric
        self.metric_params = metric_params
        self.force_dim = force_dim
        self.n_timestep_ = None
        self.n_dims_ = None

    def _validate_x_predict(self, x, check_input):
        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimensions X.ndim ({})".format(x.ndim))
        if isinstance(self.force_dim, int):
            x = np.reshape(x, [x.shape[0], self.force_dim, -1])
        if x.shape[-1] != self.n_timestep_:
            raise ValueError(
                "illegal input shape ({} != {})".format(x.shape[-1], self.n_timestep_)
            )
        if x.ndim > 2 and x.shape[1] != self.n_dims_:
            raise ValueError(
                "illegal input shape ({} != {}".format(x.shape[1], self.n_dims_)
            )
        if check_input:
            x = check_array(x, dtype=np.float64, allow_nd=True, order="C")

        if x.dtype != np.float64 or not x.flags.c_contiguous:
            x = np.ascontiguousarray(x, dtype=np.float64)
        return x

    def decision_path(self, x, check_input=True):
        check_is_fitted(self, ["tree_"])
        x = self._validate_x_predict(x, check_input)
        return self.tree_.decision_path(x)

    def apply(self, x, check_input=True):
        check_is_fitted(self, ["tree_"])
        x = self._validate_x_predict(x, check_input)
        return self.tree_.apply(x)


class ShapeletTreeRegressor(RegressorMixin, BaseShapeletTree):
    """A shapelet tree regressor.

    Attributes
    ----------

    tree_ : Tree
    """

    def __init__(
        self,
        *,
        max_depth=None,
        min_samples_split=2,
        n_shapelets=10,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="euclidean",
        metric_params=None,
        force_dim=None,
        random_state=None
    ):
        """Construct a shapelet tree regressor

        Parameters
        ----------
        max_depth : int, optional
            The maximum depth of the tree. If `None` the tree is expanded until all leaves
            are pure or until all leaves contain less than `min_samples_split` samples

        min_samples_split : int, optional
            The minimum number of samples to split an internal node

        n_shapelets : int, optional
            The number of shapelets to sample at each node.

        min_shapelet_size : float, optional
            The minimum length of a sampled shapelet expressed as a fraction, computed as
            `min(ceil(X.shape[-1] * min_shapelet_size), 2)`.

        max_shapelet_size : float, optional
            The maximum length of a sampled shapelet, expressed as a fraction, computed as
            `ceil(X.shape[-1] * max_shapelet_size)`.

        metric : {'euclidean', 'scaled_euclidean', 'scaled_dtw'}, optional
            Distance metric used to identify the best shapelet.

        metric_params : dict, optional
            Parameters for the distance measure

        force_dim : int, optional
            Force the number of dimensions.

            - If int, force_dim reshapes the input to shape (n_samples, force_dim, -1)
              for interoperability with `sklearn`.

        random_state : int or RandomState
            - If `int`, `random_state` is the seed used by the random number generator;
            - If `RandomState` instance, `random_state` is the random number generator;
            - If `None`, the random number generator is the `RandomState` instance used by `np.random`.
        """
        super(ShapeletTreeRegressor, self).__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            n_shapelets=n_shapelets,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            force_dim=force_dim,
            random_state=random_state,
        )

    def _make_tree_builder(self, x, y, sample_weight):
        random_state = check_random_state(self.random_state)
        max_shapelet_size = int(self.n_timestep_ * self.max_shapelet_size)
        min_shapelet_size = int(self.n_timestep_ * self.min_shapelet_size)
        if min_shapelet_size < 2:
            min_shapelet_size = 2

        max_depth = self.max_depth or 2 ** 31
        return RegressionShapeletTreeBuilder(
            self.n_shapelets,
            min_shapelet_size,
            max_shapelet_size,
            max_depth,
            self.min_samples_split,
            self.metric,
            self.metric_params,
            x,
            y,
            sample_weight,
            random_state,
        )

    def fit(self, X, y, sample_weight=None, check_input=True):
        """Fit a shapelet tree regressor from the training set

        Parameters
        ----------
        X : array-like of shape (n_samples, n_timesteps) or (n_samples, n_dimensions, n_timesteps)
            The training time series.

        y : array-like of shape (n_samples,)
            Target values as floating point values

        sample_weight : array-like of shape (n_samples,)
            If `None`, then samples are equally weighted. Splits that would create child
            nodes with net zero or negative weight are ignored while searching for a split
            in each node. Splits are also ignored if they would result in any single class
            carrying a negative weight in either child node.

        check_input : bool, optional
            Allow to bypass several input checking. Don't use this parameter unless you know what you do.

        Returns
        -------

        self: object
        """
        if check_input:
            X = check_array(X, dtype=np.float64, allow_nd=True, order="C")
            y = check_array(y, dtype=np.float64, ensure_2d=False)

        if X.ndim < 2 or X.ndim > 3:
            raise ValueError("illegal input dimensions")

        n_samples = X.shape[0]
        if isinstance(self.force_dim, int):
            X = np.reshape(X, [n_samples, self.force_dim, -1])

        n_timesteps = X.shape[-1]

        if X.ndim > 2:
            n_dims = X.shape[1]
        else:
            n_dims = 1

        if len(y) != n_samples:
            raise ValueError(
                "Number of labels={} does not match "
                "number of samples={}".format(len(y), n_samples)
            )

        if X.dtype != np.float64 or not X.flags.c_contiguous:
            X = np.ascontiguousarray(X, dtype=np.float64)

        if y.dtype != np.float64 or not y.flags.c_contiguous:
            y = np.ascontiguousarray(y, dtype=np.float64)

        self.n_timestep_ = n_timesteps
        self.n_dims_ = n_dims

        tree_builder = self._make_tree_builder(X, y, sample_weight)
        tree_builder.build_tree()
        self.tree_ = tree_builder.tree_
        return self

    def predict(self, x, check_input=True):
        """Predict the regression of the input samples x.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps) or (n_samples, n_dimensions, n_timesteps])
            The input time series

        check_input : bool, optional
            Allow to bypass several input checking. Don't use this parameter unless you know what you do.

        Returns
        -------

        y : ndarray of shape (n_samples,)
            The predicted classes.
        """
        check_is_fitted(self, ["tree_"])
        x = self._validate_x_predict(x, check_input)
        return self.tree_.predict(x)


class ExtraShapeletTreeRegressor(ShapeletTreeRegressor):
    """An extra shapelet tree regressor.

    Extra shapelet trees are constructed by sampling a distance threshold
    uniformly in the range [min(dist), max(dist)].

    Attributes
    ----------

    tree_ : Tree
    """

    def __init__(
        self,
        *,
        max_depth=None,
        n_shapelets=1,
        min_samples_split=2,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="euclidean",
        metric_params=None,
        force_dim=None,
        random_state=None
    ):
        """Construct a extra shapelet tree regressor

        Parameters
        ----------
        max_depth : int, optional
            The maximum depth of the tree. If `None` the tree is expanded until all leaves
            are pure or until all leaves contain less than `min_samples_split` samples

        min_samples_split : int, optional
            The minimum number of samples to split an internal node

        n_shapelets : int, optional
            The number of shapelets to sample at each node.

        min_shapelet_size : float, optional
            The minimum length of a sampled shapelet expressed as a fraction, computed as
            `min(ceil(X.shape[-1] * min_shapelet_size), 2)`.

        max_shapelet_size : float, optional
            The maximum length of a sampled shapelet, expressed as a fraction, computed as
            `ceil(X.shape[-1] * max_shapelet_size)`.

        metric : {'euclidean', 'scaled_euclidean', 'scaled_dtw'}, optional
            Distance metric used to identify the best shapelet.

        metric_params : dict, optional
            Parameters for the distance measure

        force_dim : int, optional
            Force the number of dimensions.

            - If int, force_dim reshapes the input to shape (n_samples, force_dim, -1)
              for interoperability with `sklearn`.

        random_state : int or RandomState
            - If `int`, `random_state` is the seed used by the random number generator;
            - If `RandomState` instance, `random_state` is the random number generator;
            - If `None`, the random number generator is the `RandomState` instance used by `np.random`.
        """
        super(ExtraShapeletTreeRegressor, self).__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            n_shapelets=n_shapelets,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            force_dim=force_dim,
            random_state=random_state,
        )

    def _make_tree_builder(self, x, y, sample_weight):
        random_state = check_random_state(self.random_state)
        max_shapelet_size = int(self.n_timestep_ * self.max_shapelet_size)
        min_shapelet_size = int(self.n_timestep_ * self.min_shapelet_size)
        if min_shapelet_size < 2:
            min_shapelet_size = 2

        max_depth = self.max_depth or 2 ** 31
        return ExtraRegressionShapeletTreeBuilder(
            self.n_shapelets,
            min_shapelet_size,
            max_shapelet_size,
            max_depth,
            self.min_samples_split,
            self.metric,
            self.metric_params,
            x,
            y,
            sample_weight,
            random_state,
        )


class ShapeletTreeClassifier(ClassifierMixin, BaseShapeletTree):
    """A shapelet tree classifier.

    Attributes
    ----------

    tree_ : Tree
        The tree data structure used internally

    classes_ : ndarray of shape (n_classes,)
        The class labels

    n_classes_ : int
        The number of class labels

    See Also
    --------
    ShapeletTreeRegressor : A shapelet tree regressor.
    ExtraShapeletTreeClassifier : An extra random shapelet tree classifier.
    """

    def __init__(
        self,
        max_depth=None,
        min_samples_split=2,
        n_shapelets=10,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="euclidean",
        metric_params=None,
        force_dim=None,
        random_state=None,
    ):
        """Construct a shapelet tree classifier

        Parameters
        ----------
        max_depth : int, optional
            The maximum depth of the tree. If `None` the tree is expanded until all leaves
            are pure or until all leaves contain less than `min_samples_split` samples

        min_samples_split : int, optional
            The minimum number of samples to split an internal node

        n_shapelets : int, optional
            The number of shapelets to sample at each node.

        min_shapelet_size : float, optional
            The minimum length of a sampled shapelet expressed as a fraction, computed as
            `min(ceil(X.shape[-1] * min_shapelet_size), 2)`.

        max_shapelet_size : float, optional
            The maximum length of a sampled shapelet, expressed as a fraction, computed as
            `ceil(X.shape[-1] * max_shapelet_size)`.

        metric : {'euclidean', 'scaled_euclidean', 'scaled_dtw'}, optional
            Distance metric used to identify the best shapelet.

        metric_params : dict, optional
            Parameters for the distance measure

        force_dim : int, optional
            Force the number of dimensions.

            - If int, force_dim reshapes the input to shape (n_samples, force_dim, -1)
              for interoperability with `sklearn`.

        random_state : int or RandomState
            - If `int`, `random_state` is the seed used by the random number generator;
            - If `RandomState` instance, `random_state` is the random number generator;
            - If `None`, the random number generator is the `RandomState` instance used by `np.random`.
        """
        super(ShapeletTreeClassifier, self).__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            n_shapelets=n_shapelets,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            force_dim=force_dim,
            random_state=random_state,
        )
        self.n_classes_ = None

    def _make_tree_builder(self, x, y, sample_weight):
        random_state = check_random_state(self.random_state)
        max_shapelet_size = int(self.n_timestep_ * self.max_shapelet_size)
        min_shapelet_size = int(self.n_timestep_ * self.min_shapelet_size)
        if min_shapelet_size < 2:
            min_shapelet_size = 2
        max_depth = self.max_depth or 2 ** 31
        return ClassificationShapeletTreeBuilder(
            self.n_shapelets,
            min_shapelet_size,
            max_shapelet_size,
            max_depth,
            self.min_samples_split,
            self.metric,
            self.metric_params,
            x,
            y,
            sample_weight,
            random_state,
            self.n_classes_,
        )

    def fit(self, x, y, sample_weight=None, check_input=True):
        """Fit a shapelet tree regressor from the training set

        Parameters
        ----------
        X : array-like of shape (n_samples, n_timesteps) or (n_samples, n_dimensions, n_timesteps)
            The training time series.

        y : array-like of shape (n_samples,) or (n_samples, n_classes)
            The target values (class labels) as integers

        sample_weight : array-like of shape (n_samples,)
            If `None`, then samples are equally weighted. Splits that would create child
            nodes with net zero or negative weight are ignored while searching for a split
            in each node. Splits are also ignored if they would result in any single class
            carrying a negative weight in either child node.

        check_input : bool, optional
            Allow to bypass several input checking. Don't use this parameter unless you know what you do.

        Returns
        -------

        self: object
        """
        if check_input:
            x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
            y = check_array(y, ensure_2d=False)

        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimensions")

        n_samples = x.shape[0]
        if isinstance(self.force_dim, int):
            x = np.reshape(x, [n_samples, self.force_dim, -1])

        n_timesteps = x.shape[-1]

        if x.ndim > 2:
            n_dims = x.shape[1]
        else:
            n_dims = 1

        if y.ndim == 1:
            self.classes_, y = np.unique(y, return_inverse=True)
        else:
            _, y = np.nonzero(y)
            if len(y) != n_samples:
                raise ValueError("Single label per sample expected.")
            self.classes_ = np.unique(y)

        if len(y) != n_samples:
            raise ValueError(
                "Number of labels={} does not match "
                "number of samples={}".format(len(y), n_samples)
            )

        if x.dtype != np.float64 or not x.flags.c_contiguous:
            x = np.ascontiguousarray(x, dtype=np.float64)

        if not y.flags.c_contiguous:
            y = np.ascontiguousarray(y, dtype=np.intp)

        self.n_classes_ = len(self.classes_)
        self.n_timestep_ = n_timesteps
        self.n_dims_ = n_dims

        tree_builder = self._make_tree_builder(x, y, sample_weight)
        tree_builder.build_tree()
        self.tree_ = tree_builder.tree_
        return self

    def predict(self, x, check_input=True):
        """Predict the regression of the input samples x.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps) or (n_samples, n_dimensions, n_timesteps])
            The input time series

        check_input : bool, optional
            Allow to bypass several input checking. Don't use this parameter unless you know what you do.

        Returns
        -------

        y : ndarray of shape (n_samples,)
            The predicted classes.
        """
        return self.classes_[
            np.argmax(self.predict_proba(x, check_input=check_input), axis=1)
        ]

    def predict_proba(self, x, check_input=True):
        """Predict class probabilities of the input samples X.  The predicted
        class probability is the fraction of samples of the same class
        in a leaf.

        Parameters
        ----------
        x :  array-like of shape (n_samples, n_timesteps) or (n_samples, n_dimensions, n_timesteps])
            The input time series

        check_input : bool, optional
            Allow to bypass several input checking. Don't use this parameter unless you know what you do.

        Returns
        -------
        proba : ndarray of shape (n_samples, n_classes)
            The class probabilities of the input samples. The order of the classes corresponds to
            that in the attribute `classes_`
        """
        check_is_fitted(self, ["tree_"])
        x = self._validate_x_predict(x, check_input)
        return self.tree_.predict(x)


class ExtraShapeletTreeClassifier(ShapeletTreeClassifier):
    """An extra shapelet tree classifier.

    Extra shapelet trees are constructed by sampling a distance threshold
    uniformly in the range [min(dist), max(dist)].

    Attributes
    ----------

    tree_ : Tree
    """

    def __init__(
        self,
        max_depth=None,
        n_shapelets=1,
        min_samples_split=2,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="euclidean",
        metric_params=None,
        force_dim=None,
        random_state=None,
    ):
        """Construct a extra shapelet tree regressor

        Parameters
        ----------
        max_depth : int, optional
            The maximum depth of the tree. If `None` the tree is expanded until all leaves
            are pure or until all leaves contain less than `min_samples_split` samples

        min_samples_split : int, optional
            The minimum number of samples to split an internal node

        n_shapelets : int, optional
            The number of shapelets to sample at each node.

        min_shapelet_size : float, optional
            The minimum length of a sampled shapelet expressed as a fraction, computed as
            `min(ceil(X.shape[-1] * min_shapelet_size), 2)`.

        max_shapelet_size : float, optional
            The maximum length of a sampled shapelet, expressed as a fraction, computed as
            `ceil(X.shape[-1] * max_shapelet_size)`.

        metric : {'euclidean', 'scaled_euclidean', 'scaled_dtw'}, optional
            Distance metric used to identify the best shapelet.

        metric_params : dict, optional
            Parameters for the distance measure

        force_dim : int, optional
            Force the number of dimensions.

            - If int, force_dim reshapes the input to shape (n_samples, force_dim, -1)
              or interoperability with `sklearn`.

        random_state : int or RandomState
            - If `int`, `random_state` is the seed used by the random number generator;
            - If `RandomState` instance, `random_state` is the random number generator;
            - If `None`, the random number generator is the `RandomState` instance used by `np.random`.
        """
        super(ShapeletTreeClassifier, self).__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            n_shapelets=n_shapelets,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            force_dim=force_dim,
            random_state=random_state,
        )
        self.n_classes_ = None

    def _make_tree_builder(self, x, y, sample_weight):
        random_state = check_random_state(self.random_state)
        max_shapelet_size = int(self.n_timestep_ * self.max_shapelet_size)
        min_shapelet_size = int(self.n_timestep_ * self.min_shapelet_size)
        if min_shapelet_size < 2:
            min_shapelet_size = 2

        max_depth = self.max_depth or 2 ** 31
        return ExtraClassificationShapeletTreeBuilder(
            1,
            min_shapelet_size,
            max_shapelet_size,
            max_depth,
            self.min_samples_split,
            self.metric,
            self.metric_params,
            x,
            y,
            sample_weight,
            random_state,
            self.n_classes_,
        )



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import os
import re

import numpy as np
from sklearn.utils import deprecated

from ._repository import (
    ArffBundle,
    NpyBundle,
    Bundle,
    Repository,
    RepositoryCollection,
    JSONRepository,
)

from ._filter import make_filter

__all__ = [
    "Repository",
    "JSONRepository",
    "Bundle",
    "ArffBundle",
    "NpyBundle",
    "set_cache_dir",
    "get_repository",
    "clear_cache",
    "list_repositories",
    "list_bundles",
    "get_bundles",
    "install_repository",
    "load_dataset",
    "list_datasets",
    "load_datasets",
    "load_all_datasets",
    "load_two_lead_ecg",
    "load_synthetic_control",
    "load_gun_point",
]

_CACHE_DIR = None
_REPOSITORIES = RepositoryCollection()


def _split_repo_bundle(repo_bundle_name):
    """Split a repository bundle string of the format {repo}/{bundle}

    Parameters
    ----------
    repo_bundle_name : str
        {repo}/{bundle} name

    Returns
    -------
    repository : str
        Key of the repository

    bundle : str
        Name of the bundle

    version : str
        An optional version

    tag : str
        An optional tag
    """
    match = re.match(
        "(?:([a-zA-Z]+)/([a-zA-Z0-9\-]+))?(?::((?:\d+\.)?(?:\d+\.)?(?:\*|\d+)))?(?::([a-zA-Z\-]+))?$",
        repo_bundle_name,
    )
    if match:
        repository_name = match.group(1)
        bundle_name = match.group(2)
        version = match.group(3)
        if version:
            version = re.sub("(\d+\.\d+)((?:\.0+)*)$", "\\1", version)
        tag = match.group(4)
        return repository_name, bundle_name, version, tag
    else:
        raise ValueError("repository (%s) is not supported" % repo_bundle_name)


def set_cache_dir(cache_dir):
    """Change the global cache directory

    cache_dir : str
        The cache directory root
    """
    global _CACHE_DIR
    _CACHE_DIR = cache_dir


def _default_cache_dir():
    return _os_cache_path("wildboar") if not _CACHE_DIR else _CACHE_DIR


def _os_cache_path(dir):
    import platform

    if platform.system() == "Windows":
        cache_dir = os.path.expandvars(r"%LOCALAPPDATA%\cache")
        return os.path.join(cache_dir, dir)
    elif platform.system() == "Linux":
        cache_dir = os.environ.get("XDG_CACHE_HOME")
        if not cache_dir:
            cache_dir = ".cache"
        return os.path.join(os.path.expanduser("~"), cache_dir, dir)
    elif platform.system() == "Darwin":
        return os.path.join(os.path.expanduser("~"), "Library", "Caches", dir)
    else:
        return os.path.join(os.path.expanduser("~"), ".cache", dir)


def load_synthetic_control(merge_train_test=True):
    """Load the Synthetic_Control dataset

    See Also
    --------
    load_dataset : load a named dataset
    """
    return load_dataset(
        "SyntheticControl",
        repository="wildboar/ucr-tiny",
        merge_train_test=merge_train_test,
    )


def load_two_lead_ecg(merge_train_test=True):
    """Load the TwoLeadECG dataset

    See Also
    --------
    load_dataset : load a named dataset
    """
    return load_dataset(
        "TwoLeadECG",
        repository="wildboar/ucr-tiny",
        merge_train_test=merge_train_test,
    )


def load_gun_point(merge_train_test=True):
    """Load the GunPoint dataset

    See Also
    --------
    load_dataset : load a named dataset
    """
    return load_dataset(
        "GunPoint",
        repository="wildboar/ucr-tiny",
        merge_train_test=merge_train_test,
    )


def load_datasets(
    repository="wildboar/ucr",
    *,
    cache_dir=None,
    create_cache_dir=True,
    progress=True,
    force=False,
    filter=None,
    **kwargs
):
    """Load all datasets as a generator

    Parameters
    ----------
    repository : str
        The repository string

    progress : bool, optional
        If progress indicator is shown while downloading the repository.

    cache_dir : str, optional
        The cache directory for downloaded dataset repositories.

    create_cache_dir : bool, optional
        Create the cache directory if it does not exist.

    force : bool, optional
        Force re-download of cached repository

    filter: dict or callable, optional
        Filter the datasets

        - if callable, only yield those datasets for which the callable returns True.
          ``f(dataset, x, y) -> bool``

        - if dict, filter based on the keys and values
            - "dataset": regex matching dataset name
            - "n_samples": comparison spec
            - "n_timestep": comparison spec

        Comparison spec
            str of two parts, comparison operator (<, <=, >, >= or =) and a number, e.g., "<100", "<= 200", or ">300"

    kwargs : dict
        Optional arguments to ``load_dataset``

    Yields
    ------
    x : array-like
        Data samples

    y : array-like
        Data labels

    Examples
    --------

    >>> from wildboar.datasets import load_datasets
    >>> for dataset, (x, y) in load_datasets(repository='wildboar/ucr'):
    >>>     print(dataset, x.shape, y.shape)

    Print the names of datasets with more than 200 samples

    >>> for dataset, (x, y) in load_datasets(repository='wildboar/ucr', filter={"n_samples": ">200"}):
    >>>     print(dataset)
    """
    for dataset in list_datasets(
        repository=repository,
        cache_dir=cache_dir,
        create_cache_dir=create_cache_dir,
        progress=progress,
        force=force,
    ):
        x, y = load_dataset(dataset, repository=repository, **kwargs)
        if filter is None:
            yield dataset, (x, y)
        elif hasattr(filter, "__call__"):
            if filter(dataset, x, y):
                yield dataset, (x, y)
        else:
            if make_filter(filter)(dataset, x, y):
                yield dataset, (x, y)


@deprecated(
    "the function datasets.load_all_datasets has been deprecated in "
    "1.0.3 and will be removed in in 1.0.5."
)
def load_all_datasets(
    repository="wildboar/ucr",
    *,
    cache_dir=None,
    create_cache_dir=True,
    progress=True,
    force=False,
    **kwargs
):
    return load_datasets(
        repository,
        cache_dir=cache_dir,
        create_cache_dir=create_cache_dir,
        progress=progress,
        force=force,
        **kwargs
    )


def load_dataset(
    name,
    *,
    repository="wildboar/ucr",
    dtype=None,
    contiguous=True,
    merge_train_test=True,
    cache_dir=None,
    create_cache_dir=True,
    progress=True,
    force=False
):
    """Load a dataset from a repository

    Parameters
    ----------
    name : str
        The name of the dataset to load.

    repository : str, optional
        The data repository formatted as {repository}/{bundle}[:{version}][:{tag}]

    dtype : dtype, optional
        The data type of the returned data

    contiguous : bool, optional
        Ensure that the returned dataset is memory contiguous.

    merge_train_test : bool, optional
        Merge the existing training and testing partitions.

    progress: bool, optional
        Show a progress bar while downloading a bundle.

    cache_dir: str, optional
        The directory where downloaded files are cached

    create_cache_dir: bool, optional
        Create cache directory if missing (default=True)

    force : bool, optional
        Force re-download of already cached bundle

        ..versionadded :: 1.0.4

    Returns
    -------
    x : ndarray
        The data samples

    y : ndarray
        The labels

    x_train : ndarray, optional
        The training samples if ``merge_train_test=False``

    x_test : ndarray, optional
        The testing samples if ``merge_train_test=False``

    y_train : ndarray, optional
        The training labels if ``merge_train_test=False``

    y_test : ndarray, optional
        The testing labels if ``merge_train_test=False``

    Examples
    --------

    Load a dataset from the default repository

    >>> x, y = load_dataset("SyntheticControl")

    or if original training and testing splits are to be preserved

    >>> x_train, x_test, y_train, y_test = load_dataset("SyntheticControl", merge_train_test=False)

    or for a specific version of the dataset

    >>> x_train, x_test, y_train, y_test = load_dataset("Wafer", repository='wildboar/ucr-tiny:1.0')
    """
    (
        repository_name,
        bundle_name,
        bundle_version,
        bundle_tag,
    ) = _split_repo_bundle(repository)
    dtype = dtype or np.float64
    cache_dir = cache_dir or _default_cache_dir()
    repository = get_repository(repository_name)
    ret_val = []
    x, y, n_train_samples = repository.load_dataset(
        bundle_name,
        name,
        version=bundle_version,
        tag=bundle_tag,
        dtype=dtype,
        cache_dir=cache_dir,
        force=force,
        create_cache_dir=create_cache_dir,
        progress=progress,
    )

    if merge_train_test:
        ret_val.append(x)
        ret_val.append(y)
    else:
        if n_train_samples == x.shape[0]:
            raise ValueError("found no test parts. Set merge_train_test=True.")

        ret_val.append(x[:n_train_samples])
        ret_val.append(x[n_train_samples:])
        ret_val.append(y[:n_train_samples])
        ret_val.append(y[n_train_samples:])

    if contiguous:
        return [np.ascontiguousarray(x) for x in ret_val]
    else:
        return ret_val


def list_datasets(
    repository="wildboar/ucr",
    *,
    cache_dir=None,
    create_cache_dir=True,
    progress=True,
    force=False
):
    """List the datasets in the repository

    Parameters
    ----------
    repository : str or Bundle, optional
        The data repository

        - if str load a named bundle, format {repository}/{bundle}

    progress: bool, optional
        Show a progress bar while downloading a bundle.

    cache_dir: str, optional
        The directory where downloaded files are cached (default='wildboar_cache')

    create_cache_dir: bool, optional
        Create cache directory if missing (default=True)

    force : bool, optional
        Force re-download of cached bundle

    Returns
    -------
        dataset : set
            A set of dataset names
    """
    (
        repository_name,
        bundle_name,
        bundle_version,
        bundle_tag,
    ) = _split_repo_bundle(repository)
    cache_dir = cache_dir or _default_cache_dir()
    repository = get_repository(repository_name)
    return repository.list_datasets(
        bundle_name,
        version=bundle_version,
        tag=bundle_tag,
        cache_dir=cache_dir,
        create_cache_dir=create_cache_dir,
        progress=progress,
        force=force,
    )


def clear_cache(repository=None, *, cache_dir=None, keep_last_version=True):
    """Clear the cache by deleting cached datasets

    Parameters
    ----------
    repository : str, optional
        The name of the repository to clear cache.

        - if None, clear cache of all repositories

    cache_dir : str, optional
        The cache directory

    keep_last_version : bool, optional
        If true, keep the latest version of each repository.
    """
    cache_dir = cache_dir or _default_cache_dir()
    if repository is None:
        for repo in _REPOSITORIES:
            repo.clear_cache(cache_dir=cache_dir, keep_last_version=keep_last_version)
    else:
        get_repository(repository).clear_cache(
            cache_dir=cache_dir, keep_last_version=keep_last_version
        )


def get_repository(repository):
    """Get repository by name

    Parameters
    ----------
    repository : str
        Repository name

    Returns
    -------
    repository : Repository
        A repository
    """
    if repository in _REPOSITORIES:
        return _REPOSITORIES[repository]
    else:
        raise ValueError("repository (%s) does not exist" % repository)


def install_repository(repository):
    """Install repository

    Parameters
    ----------
    repository : str or Repository
        A repository
    """
    if isinstance(repository, str):
        repository = JSONRepository(repository)
    _REPOSITORIES.append(repository)


def get_bundles(repository):
    """Get all bundles in the repository

    Parameters
    ----------
    repository : str
        Name of the repository

    Returns
    -------
    dict : A dict of key Bundle pairs
    """
    if repository in _REPOSITORIES:
        return _REPOSITORIES[repository].get_bundles()
    else:
        raise ValueError("repository (%s) does not exist.")


def list_bundles(repository):
    """Get a list of all bundle names in the specified repository.

    Parameters
    ----------
    repository : str
        The name of the repository

    Returns
    -------
    bundle : str
        The name of the bundle
    """
    return [key for key, bundle in get_bundles(repository).items()]


def list_repositories():
    """List the key of all installed repositories"""
    return [repo.name for repo in _REPOSITORIES]


# Install the default 'wildboar' repository
install_repository("https://isaksamsten.github.io/wildboar-datasets/1.0/repo.json")



import numpy as np
from wildboar.datasets import load_dataset
from wildboar.explain.counterfactual import counterfactuals
from sklearn.neighbors import KNeighborsClassifier

x_train, x_test, y_train, y_test = load_dataset(
    "GunPoint", repository="wildboar/ucr", merge_train_test=False
)

clf = KNeighborsClassifier(n_neighbors=1, metric="euclidean")
clf.fit(x_train, y_train)

x_counter, success, scores = counterfactuals(
    clf, x_test, y_test[::-1], random_state=123, scoring="euclidean"
)

print(scores)
print(np.sum(success))
print(np.sum(clf.predict(x_counter) == y_test[::-1]) / y_test.shape[0])



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import math
from abc import ABCMeta
import warnings

import numpy as np
from sklearn.utils import check_random_state
from sklearn.model_selection._split import _build_repr
from sklearn.utils.validation import check_array


class RepeatedOutlierSplit(metaclass=ABCMeta):
    """Repeated random outlier cross-validator

    Yields indicies that split the dataset into training and test sets.

    Note
    ----
    Contrary to other cross-validation strategies, the random outlier
    cross-validator does not ensure that all folds will be different.
    Instead, the inlier samples are shuffled and new outlier samples
    are inserted in the training and test sets repeatedly.
    """

    def __init__(
        self,
        n_splits=None,
        *,
        test_size=0.2,
        n_outlier=0.05,
        shuffle=True,
        random_state=None,
    ):
        """Construct a new cross-validator

        Parameters
        ----------
        n_splits : int, optional
            The maximum number of splits.

            - if None, the number of splits is determined by the number of
              outliers as, `total_n_outliers/(n_inliers * n_outliers)`

            - if int, the number of splits is an upper-bound

        test_size : float, optional
            The size of the test set.

        n_outlier : float, optional
            The fraction of outliers in the training and test sets.

        shuffle : bool, optional
            Shuffle the training indicies in each iteration.

        random_state : int or RandomState, optional
            The psudo-random number generator
        """
        self.n_outliers = n_outlier
        self.test_size = test_size
        self.n_splits = n_splits
        self.shuffle = shuffle
        self.random_state = random_state

    def get_n_splits(self, X, y, groups=None):
        """Returns the number of splitting iterations in the cross-validator
        Parameters
        ----------
        X : object
            The samples
        y : object
            The labels
        groups : object
            Always ignored, exists for compatibility.
        Returns
        -------
        n_splits : int
            Returns the number of splitting iterations in the cross-validator.
        """
        outlier_index = (y == -1).nonzero()[0]
        inlier_index = (y == 1).nonzero()[0]
        total_n_outliers = len(outlier_index)
        n_outliers = min(
            total_n_outliers, math.ceil(inlier_index.shape[0] * self.n_outliers)
        )
        if self.n_splits is None:
            n_splits = total_n_outliers // n_outliers
        else:
            n_splits = min(self.n_splits, total_n_outliers // n_outliers)
        return n_splits

    def split(self, x, y, groups=None):
        """Return training and test indicies

        Parameters
        ----------
        x : object
            Always ignored, exists for compatibility.
        y : object
            The labels
        groups : object, optional
            Always ignored, exists for compatibility.

        Yields
        -------
        train_idx, test_idx : ndarray
            The training and test indicies
        """
        y = check_array(y, ensure_2d=False)
        random_state = check_random_state(self.random_state)
        outlier_index = (y == -1).nonzero()[0]
        inlier_index = (y == 1).nonzero()[0]
        total_n_outliers = len(outlier_index)
        n_inliers = inlier_index.shape[0]
        n_outliers = min(
            total_n_outliers, math.ceil(inlier_index.shape[0] * self.n_outliers)
        )
        if n_outliers < 2:
            warnings.warn("to few outliers", UserWarning)

        if self.n_splits is None:
            n_splits = total_n_outliers // n_outliers
        else:
            n_splits = min(self.n_splits, total_n_outliers // n_outliers)
        random_state.shuffle(outlier_index)

        for i in range(n_splits):
            if self.shuffle:
                random_state.shuffle(inlier_index)
            outlier_index_sample = outlier_index[
                (i * n_outliers) : (i * n_outliers + n_outliers)
            ]
            outlier_test_index = math.ceil(n_outliers * self.test_size)
            inlier_test_index = math.ceil(n_inliers * self.test_size)

            yield (
                np.concatenate(
                    [
                        inlier_index[inlier_test_index:],  # last elements
                        outlier_index_sample[outlier_test_index:],
                    ]
                ),
                np.concatenate(
                    [
                        inlier_index[:inlier_test_index],  # first elements
                        outlier_index_sample[:outlier_test_index],
                    ]
                ),
            )

    def __repr__(self):
        return _build_repr(self)



import numpy as np
from sklearn.model_selection import cross_val_score

from wildboar import datasets
from wildboar.ensemble import ExtraShapeletTreesClassifier, ShapeletForestClassifier

x, y = datasets.load_gun_point()
extra = ExtraShapeletTreesClassifier(
    n_estimators=100, n_jobs=16, metric="scaled_euclidean"
)
rsf = ShapeletForestClassifier(n_estimators=100, n_jobs=16, metric="scaled_euclidean")
score = cross_val_score(rsf, x, y, cv=10)
print("RSF", np.mean(score))
score = cross_val_score(extra, x, y, cv=10)
print("Extra", np.mean(score))



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

import numpy as np
cimport numpy as np

from libc.stdlib cimport malloc

from libc.math cimport sqrt
from libc.math cimport NAN
from libc.stdlib cimport free

from . import _dtw_distance
from . import _euclidean_distance

from ._distance cimport TSCopy
from ._distance cimport DistanceMeasure
from ._distance cimport ts_database_new
from ._distance cimport TSDatabase

from .._utils import check_array_fast

from sklearn.utils import check_array

_DISTANCE_MEASURE = {
    'euclidean': _euclidean_distance.EuclideanDistance,
    'dtw': _dtw_distance.DtwDistance,
    'scaled_euclidean': _euclidean_distance.ScaledEuclideanDistance,
    'scaled_dtw': _dtw_distance.ScaledDtwDistance,
}


cdef int ts_copy_init(
    TSCopy *ts_copy,
    Py_ssize_t dim,
    Py_ssize_t length,
    double mean,
    double std,
) nogil:
    ts_copy.dim = dim
    ts_copy.length = length
    ts_copy.mean = mean
    ts_copy.std = std
    ts_copy.data = <double*> malloc(sizeof(double) * length)
    if ts_copy.data == NULL:
        return -1


cdef void ts_copy_free(TSCopy *ts_copy) nogil:
    if ts_copy != NULL and ts_copy.data != NULL:
        free(ts_copy[0].data)


cdef void ts_view_init(TSView *ts_view) nogil:
    """Initialize  a shapelet info struct """
    ts_view.start = 0
    ts_view.length = 0
    ts_view.dim = 0
    ts_view.mean = NAN
    ts_view.std = NAN
    ts_view.index = 0


cdef void ts_view_free(TSView *ts_view) nogil:
    """Free the `extra` payload of a shapelet info if needed """
    if ts_view.extra != NULL:
        free(ts_view.extra)
        ts_view.extra = NULL


cdef int _ts_view_update_statistics(TSView *ts_view, const TSDatabase *td) nogil:
    """Update the mean and standard deviation of a shapelet info struct """
    cdef Py_ssize_t shapelet_offset = (
            ts_view.index * td.sample_stride +
            ts_view.dim * td.dim_stride +
            ts_view.start * td.timestep_stride
    )
    cdef double ex = 0
    cdef double ex2 = 0
    cdef Py_ssize_t i
    for i in range(ts_view.length):
        current_value = td.data[shapelet_offset + i * td.timestep_stride]
        ex += current_value
        ex2 += current_value ** 2

    ts_view.mean = ex / ts_view.length
    ts_view.std = sqrt(ex2 / ts_view.length - ts_view.mean * ts_view.mean)
    return 0


cdef TSDatabase ts_database_new(np.ndarray data):
    """Construct a new time series database from a ndarray """
    data = check_array_fast(data, allow_nd=True)
    if data.ndim < 2 or data.ndim > 3:
        raise ValueError("ndim {0} < 2 or {0} > 3".format(data.ndim))

    cdef TSDatabase sd
    sd.n_samples = <Py_ssize_t> data.shape[0]
    sd.n_timestep = <Py_ssize_t> data.shape[data.ndim - 1]
    sd.data = <double*> data.data
    sd.sample_stride = <Py_ssize_t> data.strides[0] / <Py_ssize_t> data.itemsize
    sd.timestep_stride = (<Py_ssize_t> data.strides[data.ndim - 1] / <Py_ssize_t> data.itemsize)

    if data.ndim == 3:
        sd.n_dims = <Py_ssize_t> data.shape[data.ndim - 2]
        sd.dim_stride = (<Py_ssize_t> data.strides[data.ndim - 2] / <Py_ssize_t> data.itemsize)
    else:
        sd.n_dims = 1
        sd.dim_stride = 0

    return sd


cdef class DistanceMeasure:
    """A distance measure can compute the distance between time series and
    shapelets """

    def __cinit__(self, Py_ssize_t n_timestep, *args, **kvargs):
        self.n_timestep = n_timestep


    def __reduce__(self):
        return self.__class__, (self.n_timestep,)


    cdef int init(self, TSDatabase *td) nogil:
        return 0


    cdef void ts_view_sub_distances(
        self,
        TSView *ts_view,
        TSDatabase *td,
        Py_ssize_t *samples,
        double *distances,
        Py_ssize_t n_samples,
    ) nogil:
        """ Compute the distance between the shapelet all samples in `samples`

        :param ts_view: information about the shapelet
        :param td: the time series database
        :param samples: array of length `n_samples` samples to compute
        the distance to
        :param distances: array to store the distances. The the length
        of distances >= `n_samples`, the `i = 0,...,n_samples`
        position stores the distance between the i:th sample (in
        `samples`) and `s` [out param]
        :param n_samples: the number of samples 
        """
        cdef Py_ssize_t p
        for p in range(n_samples):
            distances[p] = self.ts_view_sub_distance(ts_view, td, samples[p])


    cdef int init_ts_view(
        self,
        TSDatabase *_td,
        TSView *ts_view,
        Py_ssize_t index,
        Py_ssize_t start,
        Py_ssize_t length,
        Py_ssize_t dim,
    ) nogil:
        """Set the parameters of the view

        :param _td: shapelet database
        :param ts_view: [out param] 
        :param index: the index of the sample in `td`
        :param start: the start position of the subsequence
        :param length: the length of the subsequence
        :param dim: the dimension of the subsequence
        :return non-negative on success
        """
        ts_view.index = index
        ts_view.dim = dim
        ts_view.start = start
        ts_view.length = length
        ts_view.mean = NAN
        ts_view.std = NAN
        ts_view.extra = NULL
        return 0


    cdef int init_ts_copy_from_ndarray(
        self,
        TSCopy *ts_copy,
        np.ndarray arr,
        Py_ssize_t dim,
    ):
        ts_copy.dim = dim
        ts_copy.length = arr.shape[0]
        ts_copy.mean = NAN
        ts_copy.std = NAN
        ts_copy.data = <double*> malloc(ts_copy.length * sizeof(double))
        if ts_copy.data == NULL:
            return -1

        cdef Py_ssize_t i
        for i in range(ts_copy[0].length):
            ts_copy[0].data[i] = arr[i]
        return 0


    cdef int init_ts_copy(
        self,
        TSCopy *ts_copy,
        TSView *ts_view,
        TSDatabase *td,
    ) nogil:
        ts_copy_init(ts_copy, ts_view.dim, ts_view.length, ts_view.mean, ts_view.std)
        ts_copy.ts_start = ts_view.start
        ts_copy.ts_index = ts_view.index
        cdef double *data = ts_copy.data
        cdef Py_ssize_t tc_offset = (
            ts_view.index * td.sample_stride +
            ts_view.start * td.timestep_stride +
            ts_view.dim * td.dim_stride
        )

        cdef Py_ssize_t i
        cdef Py_ssize_t p

        for i in range(ts_view.length):
            p = tc_offset + td.timestep_stride * i
            data[i] = td.data[p]

        return 0


    cdef double ts_view_sub_distance(
        self,
        TSView *ts_view,
        TSDatabase *td,
        Py_ssize_t t_index,
    ) nogil:
        """Return the minimum distance
        
        Parameters
        ----------
        ts_view : the subsequence
         
        td : time series database
        
        t_index : sample to compare

        Returns
        -------
        distance : the distance
        """
        with gil:
            raise NotImplementedError()


    cdef double ts_copy_sub_distance(
        self,
        TSCopy *ts_view,
        TSDatabase *td,
        Py_ssize_t t_index,
        Py_ssize_t *return_index=NULL,
    ) nogil:
        """Return the distance between `s` and the sample specified by
        `t_index` in `td` setting the index of the best matching
        position to `return_index` unless `return_index == NULL`

        :param ts_view: the shapelet

        :param td: the time series database

        :param t_index: the sample index

        :param return_index: (out) the index of the best matching position
        """
        with gil:
            raise NotImplementedError()

    cdef int ts_copy_sub_matches(
        self,
        TSCopy *s_ptr,
        TSDatabase *td_ptr,
        Py_ssize_t t_index,
        double threshold,
        Py_ssize_t** matches,
        double** distances,
        Py_ssize_t *n_matches,
    ) nogil except -1:
        """Compute the matches for `s` in the sample `t_index` in `td` where
        the distance threshold is below `threshold`, storing the
        matching starting positions in `matches` and distance (<
        `threshold`) in `distances` with `n_matches` storing the
        number of successful matches.

        Note:

        - `matches` will be allocated and must be freed by the caller
        - `distances` will be allocated and must be freed by the caller

        :param s_ptr: the shapelet

        :param td_ptr: the time series database

        :param t_index: the sample

        :param threshold: the minimum distance to consider a match

        :param matches: (out) array of matching locations

        :param distances: (out) array of distance at the matching
        location (< `threshold`)

        :param n_matches: (out) the number of matches
        """
        with gil:
            raise NotImplementedError()


    cdef double ts_copy_distance(self, TSCopy *s, TSDatabase *td, Py_ssize_t t_index) nogil:
        return self.ts_copy_sub_distance(s, td, t_index)


    cdef bint support_unaligned(self) nogil:
        return 0


cdef class ScaledDistanceMeasure(DistanceMeasure):
    """Distance measure that uses computes the distance on mean and
    variance standardized shapelets"""

    cdef int init_ts_copy_from_ndarray(
        self,
        TSCopy *ts_copy,
        np.ndarray arr,
        Py_ssize_t dim
    ):
        cdef int err = DistanceMeasure.init_ts_copy_from_ndarray(
            self, ts_copy, arr, dim
        )
        if err == -1:
            return -1
        ts_copy.mean = np.mean(arr)
        ts_copy.std = np.std(arr)
        return 0

    cdef int init_ts_view(
        self,
        TSDatabase *td,
        TSView *ts_view,
        Py_ssize_t index,
        Py_ssize_t start,
        Py_ssize_t length,
        Py_ssize_t dim,
    ) nogil:
        DistanceMeasure.init_ts_view(self, td, ts_view, index, start, length, dim)
        _ts_view_update_statistics(ts_view, td)
        return -1


cdef class FuncDistanceMeasure(DistanceMeasure):
    cdef object func
    cdef np.ndarray x_buffer
    cdef np.ndarray y_buffer
    cdef bint _support_unaligned

    def __cinit__(self, Py_ssize_t n_timestep, object func, bint support_unaligned=False):
        self.n_timestep = n_timestep
        self.func = func
        self.x_buffer = np.empty(n_timestep, dtype=np.float64)
        self.y_buffer = np.empty(n_timestep, dtype=np.float64)
        self._support_unaligned = support_unaligned


    def __reduce__(self):
        return self.__class__, (self.n_timestep, self.func)


    cdef double ts_copy_sub_distance(
        self,
        TSCopy *ts_copy,
        TSDatabase *td,
        Py_ssize_t t_index,
        Py_ssize_t *return_index=NULL,
    ) nogil:
        cdef Py_ssize_t i
        cdef Py_ssize_t sample_offset = (t_index * td.sample_stride + ts_copy.dim * td.dim_stride)
        with gil:
            for i in range(td.n_timestep):
                if i < ts_copy.length:
                    self.x_buffer[i] = ts_copy.data[i]
                self.y_buffer[i] = td.data[sample_offset + td.timestep_stride * i]

            return self.func(self.x_buffer[:ts_copy.length], self.y_buffer)

    cdef double ts_view_sub_distance(
        self,
        TSView *ts_view,
        TSDatabase *td,
        Py_ssize_t t_index,
    ) nogil:
        cdef Py_ssize_t i
        cdef Py_ssize_t sample_offset = (t_index * td.sample_stride +
                                         ts_view.dim * td.dim_stride)
        cdef Py_ssize_t shapelet_offset = (ts_view.index * td.sample_stride +
                                           ts_view.dim * td.dim_stride +
                                           ts_view.start * td.timestep_stride)
        with gil:
            for i in range(td.n_timestep):
                if i < ts_view.length:
                    self.x_buffer[i] = td.data[shapelet_offset + td.timestep_stride * i]
                self.y_buffer[i] = td.data[sample_offset + td.timestep_stride * i]

            return self.func(self.x_buffer[:ts_view.length], self.y_buffer)

    cdef double ts_copy_distance(
        self,
        TSCopy *ts_copy,
        TSDatabase *td,
        Py_ssize_t t_index,
    ) nogil:
        return self.ts_copy_sub_distance(ts_copy, td, t_index)

    cdef bint support_unaligned(self) nogil:
        return self._support_unaligned


def _validate_shapelet(shapelet):
    cdef np.ndarray s = check_array(
        shapelet, ensure_2d=False, dtype=np.float64, order="c")
    if s.ndim > 1:
        raise ValueError("only 1d shapelets allowed")

    if not s.flags.c_contiguous:
        s = np.ascontiguousarray(s, dtype=np.float64)
    return s


def _validate_data(data):
    cdef np.ndarray x = check_array(
        data, ensure_2d=False, allow_nd=True, dtype=np.float64, order="c")
    if x.ndim == 1:
        x = x.reshape(-1, x.shape[0])

    if not x.flags.c_contiguous:
        x = np.ascontiguousarray(x, dtype=np.float64)
    return x


def _check_sample(sample, n_samples):
    if sample < 0 or sample >= n_samples:
        raise ValueError("illegal sample {}".format(sample))


def _check_dim(dim, ndims):
    if dim < 0 or dim >= ndims:
        raise ValueError("illegal dimension {}".format(dim))


cdef np.ndarray _new_match_array(Py_ssize_t *matches, Py_ssize_t n_matches):
    if n_matches > 0:
        match_array = np.empty(n_matches, dtype=np.intp)
        for i in range(n_matches):
            match_array[i] = matches[i]
        return match_array
    else:
        return None


cdef np.ndarray _new_distance_array(
        double *distances, Py_ssize_t n_matches):
    if n_matches > 0:
        dist_array = np.empty(n_matches, dtype=np.float64)
        for i in range(n_matches):
            dist_array[i] = distances[i]
        return dist_array
    else:
        return None


cdef DistanceMeasure new_distance_measure(
    TSDatabase *td,
    object metric,
    dict metric_params=None,
):
    """Create a new distance measure

    Parameters
    ----------
    metric : str or callable
        A metric name or callable

    td : TSDatabase
        Number of maximum number of timesteps in the database

    metric_params : dict, optional
        Parameters to the metric

    Returns
    -------
    distance_measure : a distance measure instance
    """
    cdef DistanceMeasure distance_measure
    metric_params = metric_params or {}
    if isinstance(metric, str):
        if metric in _DISTANCE_MEASURE:
            distance_measure = _DISTANCE_MEASURE[metric](td.n_timestep, **metric_params)
        else:
            raise ValueError("metric (%s) is not supported" % metric)
    elif hasattr(metric, "__call__"):
        distance_measure = FuncDistanceMeasure(td.n_timestep, metric)
    else:
        raise ValueError("unknown metric, got %r" % metric)
    distance_measure.init(td)
    return distance_measure


def distance(shapelet, data, dim=0, sample=None, metric="euclidean", metric_params=None, subsequence_distance=True, return_index=False):
    cdef np.ndarray s = _validate_shapelet(shapelet)
    cdef np.ndarray x = _validate_data(data)
    if sample is None:
        if x.shape[0] == 1:
            sample = 0
        else:
            sample = np.arange(x.shape[0])
    cdef TSDatabase sd = ts_database_new(x)

    _check_dim(dim, sd.n_dims)
    cdef double min_dist
    cdef Py_ssize_t min_index

    cdef double mean = 0
    cdef double std = 0

    cdef DistanceMeasure distance_measure = new_distance_measure(
        &sd, metric, metric_params
    )

    if (
            not subsequence_distance and
            not distance_measure.support_unaligned() and
            s.shape[0] != sd.n_timestep
    ):
        raise ValueError(
            "x.shape[0] != y.shape[-1], got %r, %r" % (s.shape[0], sd.n_timestep)
        )

    if not distance_measure.support_unaligned() and s.shape[0] > sd.n_timestep:
        raise ValueError(
            "x.shape[0] > y.shape[-1], got %r, %r" % (s.shape[0], sd.n_timestep)
        )


    cdef TSCopy shape
    distance_measure.init_ts_copy_from_ndarray(&shape, s, dim)
    if isinstance(sample, int):
        if subsequence_distance:
            min_dist = distance_measure.ts_copy_sub_distance(
                &shape, &sd, sample, &min_index)
        else:
            min_dist = distance_measure.ts_copy_distance(
                &shape, &sd, sample
            )
            min_index = 0

        if return_index:
            return min_dist, min_index
        else:
            return min_dist
    else:  # assume an `array_like` object for `samples`
        samples = check_array(sample, ensure_2d=False, dtype=np.int)
        dist = []
        ind = []
        for i in samples:
            if subsequence_distance:
                min_dist = distance_measure.ts_copy_sub_distance(
                    &shape, &sd, i, &min_index
                )
            else:
                min_dist = distance_measure.ts_copy_distance(
                    &shape, &sd, sample
                )
                min_index = 0

            dist.append(min_dist)
            ind.append(min_index)

        if return_index:
            return np.array(dist), np.array(ind)
        else:
            return np.array(dist)


def matches(shapelet, X, threshold, dim=0, sample=None, metric="euclidean", metric_params=None, return_distance=False):
    cdef np.ndarray s = _validate_shapelet(shapelet)
    cdef np.ndarray x = _validate_data(X)
    _check_dim(dim, x.ndim)
    if sample is None:
        if x.shape[0] == 1:
            sample = 0
        else:
            sample = np.arange(x.shape[0])

    cdef TSDatabase sd = ts_database_new(x)

    cdef Py_ssize_t *matches
    cdef double *distances
    cdef Py_ssize_t n_matches

    cdef DistanceMeasure distance_measure = new_distance_measure(
        &sd, metric, metric_params
    )
    cdef TSCopy shape
    distance_measure.init_ts_copy_from_ndarray(&shape, s, dim)
    cdef Py_ssize_t i
    if isinstance(sample, int):
        _check_sample(sample, sd.n_samples)
        distance_measure.ts_copy_sub_matches(
            &shape, &sd, sample, threshold, &matches, &distances, &n_matches)

        match_array = _new_match_array(matches, n_matches)
        distance_array = _new_distance_array(distances, n_matches)
        free(distances)
        free(matches)

        if return_distance:
            return distance_array, match_array
        else:
            return match_array
    else:
        samples = check_array(sample, ensure_2d=False, dtype=np.int)
        match_list = []
        distance_list = []
        for i in samples:
            _check_sample(i, sd.n_samples)
            distance_measure.ts_copy_sub_matches(&shape, &sd, i, threshold, &matches, &distances, &n_matches)
            match_array = _new_match_array(matches, n_matches)
            distance_array = _new_distance_array(distances, n_matches)
            match_list.append(match_array)
            distance_list.append(distance_array)
            free(matches)
            free(distances)

        if return_distance:
            return distance_list, match_list
        else:
            return match_list



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten
from __future__ import print_function

import numpy as np
cimport numpy as np

from libc.math cimport log2
from libc.math cimport sqrt
from libc.stdlib cimport realloc

cdef class RollingVariance:
    def __cinit__(self):
        self._m = 0.0
        self._n_samples = 0.0
        self._s = 0.0

    cdef void _reset(self) nogil:
        self._m = 0.0
        self._n_samples = 0.0
        self._s = 0.0
        self._sum = 0.0

    def reset(self):
        self._reset()

    cdef void _add(self, double weight, double value) nogil:
        cdef double next_m
        self._n_samples += weight
        next_m = self._m + (value - self._m) / self._n_samples
        self._s += (value - self._m) * (value - next_m)
        self._m = next_m
        self._sum += weight * value

    def add(self, value, weight=1.0):
        self._add(weight, value)

    cdef void _remove(self, double weight, double value) nogil:
        cdef double old_m
        if self._n_samples == 1.0:
            self._n_samples = 0.0
            self._m = 0.0
            self._s = 0.0
        else:
            old_m = (self._n_samples * self._m - value) / (self._n_samples - weight)
            self._s -= (value - self._m) * (value - old_m)
            self._m = old_m
            self._n_samples -= weight
        self._sum -= weight * value

    def remove(self, value, weight=1.0):
        self._remove(weight, value)

    @property
    def n_samples(self):
        return self._n_samples

    @property
    def sum(self):
        return self._sum

    cdef double _mean(self) nogil:
        return self._m

    @property
    def mean(self):
        return self._mean()

    cdef double _variance(self) nogil:
        return 0.0 if self._n_samples <= 1 else self._s / self._n_samples

    @property
    def variance(self):
        return self._variance()

# For debugging
def print_tree(o, indent=1):
    if o.is_classification_leaf:
        print("-" * indent, "leaf: ")
        print("-" * indent, " proba: ", o.proba)
    elif o.is_regression_leaf:
        print("-" * indent, "leaf: ")
        print("-" * indent, " avg: ", o.regr_val)
    else:
        print("-" * indent, "branch:")
        print("-" * indent, " shapelet: ", o.shapelet.array)
        print("-" * indent, " threshold: ", o.threshold)
        print("-" * indent, " left:", end="\n")
        print_tree(o.left, indent + 1)
        print("-" * indent, " right:", end="\n")
        print_tree(o.right, indent + 1)

# For debugging
cdef void print_c_array_d(object name, double *arr, Py_ssize_t length):
    print(name, end=": ")
    for i in range(length):
        print(arr[i], end=" ")
    print()

# For debugging
cdef void print_c_array_i(object name, Py_ssize_t *arr, Py_ssize_t length):
    print(name, end=": ")
    for i in range(length):
        print(arr[i], end=" ")
    print()

cdef Py_ssize_t label_distribution(const Py_ssize_t *samples,
                               const double *sample_weights,
                               Py_ssize_t start,
                               Py_ssize_t end,
                               const Py_ssize_t *labels,
                               Py_ssize_t label_stride,
                               Py_ssize_t n_labels,
                               double *n_weighted_samples,
                               double *dist) nogil:
    """Computes the label distribution

    :param samples: the samples to include
    :param sample_weights: 
    :param start: the start position in samples
    :param end: the end position in samples
    :param labels: the labels
    :param label_stride: the stride in labels
    :param n_labels: the number labeles
    :param n_weighted_samples: (out) number of samples according to weight
    :param dist: (out) label distribution
    :return: number of classes included in the sample
    """
    cdef double sample_weight
    cdef Py_ssize_t i, j, p, n_pos

    n_pos = 0
    n_weighted_samples[0] = 0
    for i in range(start, end):
        j = samples[i]
        p = j * label_stride

        if sample_weights != NULL:
            sample_weight = sample_weights[j]
        else:
            sample_weight = 1.0

        dist[labels[p]] += sample_weight
        n_weighted_samples[0] += sample_weight

    for i in range(n_labels):
        if dist[i] > 0:
            n_pos += 1

    return n_pos

cdef inline size_t rand_r(size_t *seed) nogil:
    """Returns a pesudo-random number based on the seed.

    :param seed: the initial seed (updated)
    :return: a psudo-random number
    """
    seed[0] = seed[0] * 1103515245 + 12345
    return seed[0] % (<size_t> RAND_R_MAX + 1)

cdef inline size_t rand_int(size_t min_val, size_t max_val, size_t *seed) nogil:
    """Returns a pseudo-random number in the range [`min_val` `max_val`[

    :param min_val: the minimum value
    :param max_val: the maximum value
    :param seed: the seed (updated)
    """
    if min_val == max_val:
        return min_val
    else:
        return min_val + rand_r(seed) % (max_val - min_val)

cdef inline double rand_uniform(double low, double high, size_t *random_state) nogil:
    """Generate a random double in the range [`low` `high`[."""
    return ((high - low) * <double> rand_r(random_state) / <double> RAND_R_MAX) + low

# Implementation of introsort. Inspired by sklearn.tree
# implementation. This code is licensed under BSD3 (and not GPLv3)
#
# Including:
#  - argsort
#  - swap
#  - median3
#  - introsort
#  - sift_down
#  - heapsort


cdef inline void argsort(double *values, Py_ssize_t *samples, Py_ssize_t n) nogil:
    if n == 0:
        return
    cdef Py_ssize_t maxd = 2 * <Py_ssize_t> log2(n)
    introsort(values, samples, n, maxd)

cdef inline void swap(double *values, Py_ssize_t *samples,
                      Py_ssize_t i, Py_ssize_t j) nogil:
    values[i], values[j] = values[j], values[i]
    samples[i], samples[j] = samples[j], samples[i]

cdef inline double median3(double *values, Py_ssize_t n) nogil:
    cdef double a = values[0]
    cdef double b = values[n / 2]
    cdef double c = values[n - 1]
    if a < b:
        if b < c:
            return b
        elif a < c:
            return c
        else:
            return a
    elif b < c:
        if a < c:
            return a
        else:
            return c
    else:
        return b

cdef void introsort(double *values, Py_ssize_t *samples,
                    Py_ssize_t n, Py_ssize_t maxd) nogil:
    cdef double pivot, value
    cdef Py_ssize_t i, l, r

    while n > 1:
        if maxd <= 0:
            heapsort(values, samples, n)
            return
        maxd -= 1

        pivot = median3(values, n)

        i = l = 0
        r = n
        while i < r:
            value = values[i]
            if value < pivot:
                swap(values, samples, i, l)
                i += 1
                l += 1
            elif value > pivot:
                r -= 1
                swap(values, samples, i, r)
            else:
                i += 1

        introsort(values, samples, l, maxd)
        values += r
        samples += r
        n -= r

cdef inline void sift_down(double *values, Py_ssize_t *samples,
                           Py_ssize_t start, Py_ssize_t end) nogil:
    cdef Py_ssize_t child, maxind, root
    root = start
    while True:
        child = root * 2 + 1
        maxind = root
        if child < end and values[maxind] < values[child]:
            maxind = child
        if child + 1 < end and values[maxind] < values[child + 1]:
            maxind = child + 1

        if maxind == root:
            break
        else:
            swap(values, samples, root, maxind)
            root = maxind

cdef void heapsort(double *values, Py_ssize_t *samples, Py_ssize_t n) nogil:
    cdef Py_ssize_t start, end

    start = (n - 2) / 2
    end = n
    while True:
        sift_down(values, samples, start, end)
        if start == 0:
            break
        start -= 1

    end = n - 1
    while end > 0:
        swap(values, samples, 0, end)
        sift_down(values, samples, 0, end)
        end = end - 1

cdef int realloc_array(void** ptr, Py_ssize_t old_size, Py_ssize_t ptr_size, Py_ssize_t *capacity)  nogil except -1:
    cdef void *tmp = ptr[0]
    if old_size >= capacity[0]:
        capacity[0] = old_size * 2
        tmp = realloc(ptr[0], ptr_size * capacity[0])
        if tmp == NULL:
            return -1
    ptr[0] = tmp
    return 0

cdef int safe_realloc(void** ptr, Py_ssize_t new_size) nogil except -1:
    cdef void *tmp = ptr[0]
    tmp = realloc(ptr[0], new_size)
    if tmp == NULL:
        return -1

    ptr[0] = tmp
    return 0


cdef void fast_mean_std(
        Py_ssize_t offset,
        Py_ssize_t stride,
        Py_ssize_t length,
        double* data,
        double *mean,
        double* std,
) nogil:
    """Update the mean and standard deviation"""
    cdef double ex = 0
    cdef double ex2 = 0
    cdef Py_ssize_t i
    for i in range(length):
        current_value = data[offset + i * stride]
        ex += current_value
        ex2 += current_value ** 2

    mean[0] = ex / length
    std[0] = sqrt(ex2 / length - mean[0] * mean[0])


def check_array_fast(np.ndarray x, bint ensure_2d=False, bint allow_nd=False, bint c_order=True):
    """Ensure that the array is valid and with dtype=np.float64.

    Parameters
    ----------
    x : ndarray
        The array to validate.

    ensure_2d : bool, optional
        Ensure that the array has 2 dimensions.

    allow_nd : bool, optional
        Allow more than 2 dimensions. Only valid if ensure_2d=False.

    c_order : bool, optional
        Ensure that the returned array is in row-major order.

    Returns
    -------
    x : ndarray
        Either a copy or the original array validated.
    """
    if ensure_2d:
        if x.ndim != 2:
            raise ValueError("not 2d, got %rd" % x.ndim)
    else:
        if not allow_nd and x.ndim > 1:
            raise ValueError("not 1d, got %rd" % x.ndim)

    if not x.flags.c_contiguous and c_order:
        x = np.ascontiguousarray(x, dtype=np.float64)
    if not x.dtype == np.float64:
        x = x.astype(np.float64)
    return x



import matplotlib.pylab as plt
from sklearn.decomposition import PCA
from sklearn.model_selection import train_test_split
from sklearn.pipeline import make_pipeline

from wildboar.datasets import load_dataset
from wildboar.ensemble import ShapeletForestEmbedding, IsolationShapeletForest

random_state = 1234

x, y = load_dataset("CBF", repository="wildboar/outlier:easy")
x_train, x_test, y_train, y_test = train_test_split(
    x, y, test_size=0.2, random_state=random_state, stratify=y
)

metric = "euclidean"
embedding = make_pipeline(
    ShapeletForestEmbedding(
        metric=metric, random_state=random_state, sparse_output=False
    ),
    PCA(n_components=2, random_state=random_state),
)
isf = IsolationShapeletForest(
    contamination=0.05,
    metric=metric,
    random_state=random_state,
    n_jobs=-1,
)
isf.fit(x_train)
embedding.fit(x_train)
x_embedding = embedding.transform(x_test)

y_score = isf.decision_function(x_test)
y_pred = isf.predict(x)

fig, ax = plt.subplots()
mapping = ax.scatter(
    x_embedding[:, 0], x_embedding[:, 1], c=y_score, cmap="viridis", marker="x"
)
ax.scatter(
    x_embedding[y_test == -1, 0],
    x_embedding[y_test == -1, 1],
    edgecolors="red",
    linewidths=1,
    alpha=0.2,
    facecolors="None",
    s=100,
    marker="o",
)
ax.scatter(
    x_embedding[y_pred == -1, 0],
    x_embedding[y_pred == -1, 1],
    edgecolors="black",
    linewidths=1,
    alpha=0.2,
    facecolors="None",
    s=200,
    marker="o",
)
plt.tight_layout()
fig.colorbar(mapping, ax=ax, orientation="horizontal")
plt.savefig("../fig/outlier_isf.png")



import matplotlib.pylab as plt
from wildboar.datasets import load_synthetic_control


def time_series_predict():
    x, y = load_synthetic_control()
    fig, ax = plt.subplots()
    for i in [-1, 12]:
        ax.plot(x[i, :], label="%s" % str(y[i]))
    ax.legend()
    return fig


PLOT_DICT = {"time_series_predict": time_series_predict}



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten
import math
import numpy as np

from copy import deepcopy
from functools import partial

from wildboar.distance import distance, matches

from wildboar.ensemble._ensemble import BaseShapeletForestClassifier
from wildboar.explain.counterfactual.base import BaseCounterfactual

from sklearn.utils.validation import check_is_fitted, check_array

MIN_MATCHING_DISTANCE = 0.0001

euclidean_distance = partial(distance, metric="euclidean", return_index=True)


def _shapelet_transform(shapelet, x, start_index, theta):
    x_shapelet = x[start_index : (start_index + shapelet.shape[0])]
    shapelet_diff = x_shapelet - shapelet
    dist = np.linalg.norm(shapelet_diff)
    if dist == 0:
        x_shapelet = np.random.uniform(shapelet.shape)
        shapelet_diff = x_shapelet - shapelet
        dist = np.linalg.norm(shapelet_diff)

    return shapelet + shapelet_diff / dist * theta


class PredictionPaths:
    def __init__(self, classes):
        self.classes = classes
        self._paths = {c: [] for c in classes}

    def _append(self, tree):
        left = tree.left
        right = tree.right
        threshold = tree.threshold
        shapelet = tree.shapelet
        value = tree.value

        def recurse(node_id, path):
            if left[node_id] != -1:
                left_path = path.copy()
                (dim, data) = shapelet[node_id]
                left_path.append((-1, (dim - 1, data), threshold[node_id]))
                recurse(left[node_id], left_path)

                right_path = path.copy()
                right_path.append((1, (dim - 1, data), threshold[node_id]))
                recurse(right[node_id], right_path)
            else:
                self._paths[self.classes[np.argmax(value[node_id])]].append(path)

        recurse(0, [])

    def __contains__(self, item):
        return item in self._paths

    def __getitem__(self, item):
        return self._paths[item]


def _compute_cost(a, b):
    if a.ndim == 1 and b.ndim == 1:
        return np.linalg.norm(a - b)
    else:
        return np.linalg.norm(a - b, axis=1)


class ShapeletForestCounterfactual(BaseCounterfactual):
    """Counterfactual explanations for shapelet forest classifiers

    Attributes
    ----------
    paths_ : dict
        A dictionary of prediction paths per label

    Notes
    -----
    This implementation only supports the reversible algorithm described by Karlsson (2020)

    Warnings
    --------
    Only shapelet forests fit with the Euclidean distance is supported i.e., ``metric="euclidean"``

    References
    ----------
    Karlsson, I., Rebane, J., Papapetrou, P., & Gionis, A. (2020).
        Locally and globally explainable time series tweaking.
        Knowledge and Information Systems, 62(5), 1671-1700.

    Karlsson, I., Rebane, J., Papapetrou, P., & Gionis, A. (2018).
        Explainable time series tweaking via irreversible and reversible temporal
        transformations. In 2018 IEEE International Conference on Data Mining (ICDM)
    """

    def __init__(self, *, epsilon=1.0, batch_size=1, random_state=10):
        """

        Parameters
        ----------
        epsilon : float, optional
            Control the degree of change from the decision threshold

        batch_size : float, optional
            Batch size when evaluating the cost and predictions of counterfactual candidates.
            The default setting is to evaluate all counterfactual samples.

        random_state : RandomState or int, optional
            Pseudo-random number for consistency between different runs
        """
        self.epsilon = epsilon
        self.random_state = random_state
        self.batch_size = batch_size

    def fit(self, estimator):
        if not isinstance(estimator, BaseShapeletForestClassifier):
            raise ValueError("unsupported estimator, got %r" % estimator)
        check_is_fitted(estimator)
        self._estimator = deepcopy(estimator)
        self.paths_ = PredictionPaths(estimator.classes_)
        for base_estimator in self._estimator.estimators_:
            self.paths_._append(base_estimator.tree_)
        return self

    def transform(self, x, y):
        check_is_fitted(self, "paths_")
        y = check_array(y, ensure_2d=False)
        x = check_array(x)
        if len(y) != x.shape[0]:
            raise ValueError(
                "Number of labels={} does not match "
                "number of samples={}".format(len(y), x.shape[0])
            )
        counterfactuals = np.empty(x.shape)
        success = np.zeros(x.shape[0], dtype=bool)
        for i in range(x.shape[0]):
            t = self.candidates(x[i, :], y[i])
            if t is not None:
                counterfactuals[i, :] = t
                success[i] = True
            else:
                success[i] = False

        return counterfactuals, success

    def candidates(self, x, y):
        if x.ndim == 2 and x.shape[0] > 1:
            raise ValueError(
                "can only compute candidates for 1 sample, got %d" % x.shape[0]
            )
        # elif x.ndim == 1:
        #     x = x.reshape(1, -1)

        if self.epsilon < 0.0:
            raise ValueError("epsilon must be larger than 0, got %r" % self.epsilon)

        if y not in self.paths_:
            raise ValueError("unknown label, got %r" % y)

        prediction_paths = self.paths_[y]
        n_counterfactuals = len(prediction_paths)

        if isinstance(self.batch_size, float):
            if not 0.0 < self.batch_size <= 1:
                raise ValueError(
                    "batch_size should be in range (0, 1], got %r" % self.batch_size
                )
            batch_size = math.ceil(n_counterfactuals * self.batch_size)
        elif isinstance(self.batch_size, int):
            batch_size = max(0, min(self.batch_size, n_counterfactuals))
        else:
            raise ValueError("unsupported batch_size, got %r" % self.batch_size)

        counterfactuals = np.empty([n_counterfactuals, x.shape[0]])
        for i, path in enumerate(prediction_paths):
            counterfactuals[i, :] = self._path_transform(x.copy(), path)

        # Note that the cost is ordered in increasing order; hence, if a conversion is successful
        # there can exist no other successful transformation with lower cost.
        cost_sort = np.argsort(_compute_cost(counterfactuals, x))
        for i in range(0, n_counterfactuals, batch_size):
            batch_cost = cost_sort[i : min(n_counterfactuals, i + batch_size)]
            batch_counterfactuals = counterfactuals[batch_cost, :]
            batch_prediction = self._estimator.predict(batch_counterfactuals)
            batch_counterfactuals = batch_counterfactuals[batch_prediction == y]
            if batch_counterfactuals.shape[0] > 0:
                return batch_counterfactuals[0, :]

        return None

    def _path_transform(self, x, path):
        for direction, (dim, shapelet), threshold in path:
            dist, location = euclidean_distance(shapelet, x, dim=dim)
            if direction < 0:
                if dist > threshold:
                    impute_shape = _shapelet_transform(
                        shapelet, x, location, threshold - self.epsilon
                    )
                    x[location : location + len(shapelet)] = impute_shape
            elif direction > 0:
                while dist - threshold < 0:
                    impute_shape = _shapelet_transform(
                        shapelet, x, location, threshold + self.epsilon
                    )
                    x[location : location + len(shapelet)] = impute_shape
                    dist, location = euclidean_distance(shapelet, x, dim=dim)
            else:
                raise ValueError("invalid path")
        return x


# class IncrementalTreeLabelTransform(CounterfactualTransformer):
#     def _transform_single_path(self, x, path):
#         for direction, shapelet, threshold in path:
#             if direction == "<=":
#                 dist, location = distance(
#                     shapelet, x, metric="euclidean", return_index=True)
#                 if dist > threshold:
#                     impute_shape = _shapelet_transform(shapelet, x, location,
#                                                        threshold - self.epsilon)
#                     x[location:(location + len(shapelet))] = impute_shape
#             else:
#                 dist, location = distance(
#                     shapelet, x, metric="euclidean", return_index=True)
#                 while dist - threshold < 0.0001:
#                     impute_shape = _shapelet_transform(shapelet, x, location,
#                                                        threshold + self.epsilon)
#                     x[location:(location + len(shapelet))] = impute_shape
#                     dist, location = distance(
#                         shapelet, x, metric="euclidean", return_index=True)
#         return x
#
#
# def locked_iter(locked, start, end):
#     """iterates over unlocked regions. `locked must be sorterd`
#     :param locked: a list of tuples `[(start_lock, end_lock)]`
#     :param start: first index
#     :param end: last index
#     :yield: unlocked regions
#     """
#     if len(locked) == 0:
#         yield start, end
#     for i in range(0, len(locked)):
#         s, e = locked[i]
#         if i == 0:
#             start = 0
#         if s - start > 0:
#             yield start, (start + (s - start))
#         start = e
#
#         if i == len(locked) - 1 and end - start > 0:
#             yield start, end
#
#
# def in_range(i, start, end):
#     return start < i <= end
#
#
# def is_locked(start, end, locked):
#     """Return true if location is locked
#     :param pos: the position
#     :param locked: list of locked positions
#     :returns: true if locked
#     """
#     for s, e in locked:
#         if in_range(s, start, end) or in_range(e, start, end):
#             return True
#     return False
#
#
# class LockingIncrementalTreeLabelTransform(IncrementalTreeLabelTransform):
#     def _transform_single(self, x_orig):
#         path_list = self.paths_[self.to_label_]
#         best_cost = np.inf
#         best_x_prime = np.nan
#         batch_size = round(len(path_list) * self.batch_size) + 1
#         x_prime = np.empty((batch_size, x_orig.shape[0]))
#         n_paths = len(path_list)
#         i = 0
#         prune = 0
#         predictions = 0
#         while i < n_paths:
#             j = 0
#             while j < batch_size and i + j < n_paths:
#                 path = path_list[i + j]
#                 x_prime_j = self._transform_single_path(
#                     x_orig, path, best_cost)
#                 if x_prime_j is not None:
#                     x_prime[j, :] = x_prime_j
#                     j += 1
#                 else:
#                     prune += 1
#                     i += 1
#             i += j
#             if j > 0:
#                 predictions += j
#                 x_prime_pred = x_prime[:j, :]
#                 cond = self.ensemble_.predict(x_prime_pred) == self.to_label_
#                 x_prime_i = x_prime_pred[cond]
#                 if x_prime_i.shape[0] > 0:
#                     cost = self._compute_cost(x_prime_i, x_orig)
#                     argmin_cost = np.argmin(cost)
#                     min_cost = cost[argmin_cost]
#                     if min_cost < best_cost:
#                         best_cost = min_cost
#                         best_x_prime = x_prime_i[argmin_cost]
#
#         return (best_x_prime, predictions / float(len(path_list)),
#                 prune / float(len(path_list)))
#
#     def _transform_single_path(self, x_orig, path, best_cost=np.inf):
#         x_prime = x_orig.copy()
#         locked = []
#         for direction, shapelet, threshold in path:
#             if direction == "<=":
#                 dist, location = distance(
#                     shapelet, x_prime, metric="euclidean", return_index=True)
#
#                 if dist > threshold and not is_locked(
#                         location, location + len(shapelet), locked):
#                     impute_shape = _shapelet_transform(
#                         shapelet, x_prime, location, threshold - self.epsilon)
#
#                     x_prime[location:(location + len(shapelet))] = impute_shape
#                     locked.append((location, location + len(shapelet)))
#
#                     cost = self._compute_cost(x_prime, x_orig)
#                     if cost >= best_cost:
#                         return None
#
#             else:
#                 dist, location = distance(
#                     shapelet, x_prime, metric="euclidean", return_index=True)
#                 while (dist - threshold < 0.0001 and not is_locked(
#                         location, location + len(shapelet), locked)):
#                     impute_shape = _shapelet_transform(
#                         shapelet, x_prime, location, threshold + self.epsilon)
#
#                     x_prime[location:(location + len(shapelet))] = impute_shape
#                     locked.append((location, location + len(shapelet)))
#
#                     cost = self._compute_cost(x_prime, x_orig)
#                     if cost >= best_cost:
#                         return None
#
#                     dist, location = distance(
#                         shapelet,
#                         x_prime,
#                         metric="euclidean",
#                         return_index=True)
#         return x_prime



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import numpy as np
from sklearn.cluster import KMeans
from sklearn.neighbors import NearestNeighbors, KNeighborsClassifier
from sklearn.utils.validation import check_is_fitted

from .base import BaseCounterfactual


class KNeighborsCounterfactual(BaseCounterfactual):
    """Fit a counterfactual explainer to a k-nearest neighbors classifier

    Attributes
    ----------

    explainer_ : dict
        The explainer for each label

    References
    ----------

    Karlsson, I., Rebane, J., Papapetrou, P., & Gionis, A. (2020).
        Locally and globally explainable time series tweaking.
        Knowledge and Information Systems, 62(5), 1671-1700.
    """

    def __init__(self, random_state=None):
        self.random_state = random_state

    def fit(self, estimator):
        if not isinstance(estimator, KNeighborsClassifier):
            raise ValueError("not a valid estimator")
        check_is_fitted(estimator)
        if estimator.metric != "euclidean":
            raise ValueError(
                "only euclidean distance is supported, got %r" % estimator.metric
            )

        x = estimator._fit_X
        y = estimator._y
        classes = estimator.classes_
        n_clusters = x.shape[0] // estimator.n_neighbors
        kmeans = KMeans(n_clusters=n_clusters, random_state=self.random_state).fit(x)
        n_classes = len(classes)

        label_nn = {}
        for cls in classes:
            to_idx = np.nonzero(classes == cls)[0][0]
            center_majority = np.zeros([kmeans.n_clusters, n_classes])
            for cluster_label, class_label in zip(kmeans.labels_, y):
                center_majority[cluster_label, class_label] += 1

            center_prob = center_majority / np.sum(center_majority, axis=1).reshape(
                -1, 1
            )
            majority_class = center_prob[:, to_idx] > (1.0 / n_classes)
            maximum_class = (
                center_majority[:, to_idx] >= (estimator.n_neighbors // n_classes) + 1
            )
            cluster_centers = kmeans.cluster_centers_
            majority_centers = cluster_centers[majority_class & maximum_class, :]
            majority_center_nn = NearestNeighbors(n_neighbors=1, metric="euclidean")
            if majority_centers.shape[0] > 0:
                majority_center_nn.fit(majority_centers)

            label_nn[cls] = (majority_center_nn, majority_centers)
        self.explainer_ = label_nn
        return self

    def transform(self, x, y):
        check_is_fitted(self, ["explainer_"])
        x_counter = np.empty(x.shape, dtype=x.dtype)
        success = np.empty(x.shape[0], dtype=bool)
        labels = np.unique(y)
        for label in labels:
            label_indices = np.where(y == label)[0]

            nn, mc = self.explainer_[label]
            if mc.shape[0] > 0:
                closest = nn.kneighbors(x[label_indices, :], return_distance=False)
                x_counter[label_indices, :] = mc[closest[:, 0], :]
                success[label_indices] = True
            else:
                success[label_indices] = False
        return x_counter, success



import numpy as np
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from sklearn.metrics import accuracy_score
from sklearn.model_selection import train_test_split
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline

from wildboar.datasets import load_dataset
from wildboar.ensemble import ShapeletForestEmbedding, ShapeletForestClassifier

random_state = 1234

x, y = load_dataset("Car")
x_train, x_test, y_train, y_test = train_test_split(
    x, y, test_size=0.2, random_state=random_state
)

f_embedding = make_pipeline(
    ShapeletForestEmbedding(sparse_output=False, random_state=random_state),
    PCA(n_components=2, random_state=random_state),
)
f_embedding.fit(x_train)
x_embedding = f_embedding.transform(x_test)

classifiers = [
    ("Shapelet forest", ShapeletForestClassifier(random_state=random_state)),
    ("Nearest neighbors", KNeighborsClassifier()),
]

classes = np.unique(y)
n_classes = len(classes)

fig, ax = plt.subplots(
    nrows=len(classifiers),
    ncols=n_classes,
    figsize=(3 * n_classes, 6),
    sharex=True,
    sharey=True,
)
for i, (name, clf) in enumerate(classifiers):
    clf.fit(x_train, y_train)

    y_pred = clf.predict(x_test)
    probas = clf.predict_proba(x_test)

    for k in range(n_classes):
        if i == 0:
            ax[i, k].set_title("Label: %r" % classes[k])
        if k == 0:
            ax[i, k].set_ylabel(name)

        ax[i, k].scatter(
            x_embedding[:, 0],
            x_embedding[:, 1],
            c="black",
            alpha=0.2,
            marker="x",
        )
        mappable = ax[i, k].scatter(
            x_embedding[y_pred == classes[k], 0],
            x_embedding[y_pred == classes[k], 1],
            c=probas[y_pred == classes[k], k],
            marker="x",
            cmap="viridis",
        )
        ax[i, k].scatter(
            x_embedding[(y_test[y_test != y_pred] == classes[k]).nonzero()[0], 0],
            x_embedding[(y_test[y_test != y_pred] == classes[k]).nonzero()[0], 1],
            edgecolors="red",
            linewidths=2,
            alpha=0.3,
            facecolors="None",
            s=70,
            marker="o",
        )


plt.tight_layout()
fig.colorbar(mappable, ax=ax, orientation="horizontal")
plt.savefig("../fig/classification_sf_vs_nn.png")



import os
import sys

sys.path.insert(1, os.path.abspath("code/"))


if __name__ == "__main__":
    img_path = os.path.abspath("_static/img")
    import img

    if not os.path.exists(img_path):
        os.mkdir(img_path)

    img.build_all(img_path, img.TUTORIAL, subdir="tutorial")
    img.build_all(img_path, img.COUNTERFACTUALS, subdir="explain")



import numpy as np
from sklearn.utils import check_random_state

train = np.loadtxt("data/synthetic_control_TRAIN", delimiter=",")
test = np.loadtxt("data/synthetic_control_TEST", delimiter=",")

train = np.vstack([train, test])

random = check_random_state(123)
y = train[:, 0].astype(np.float64)
one = (y == 1) | (y == 2) | (y == 3)
y[one] = random.randn(np.sum(one))
y[~one] = 10 + random.randn(np.sum(~one))

# for k in np.unique(y):
#     y[y == k] = random_state.randn(y[y == k].shape[0]) + k
y = np.ascontiguousarray(y)
x = np.ascontiguousarray(train[:, 1:].astype(np.float64))

from sklearn.model_selection import train_test_split

x_train, x_test, y_train, y_test = train_test_split(
    x, y, test_size=0.3, random_state=123
)

print(x_train.shape)

from wildboar.ensemble import ExtraShapeletTreesRegressor

b = ExtraShapeletTreesRegressor(
    random_state=123,
    n_estimators=100,
    n_jobs=16,
    min_shapelet_size=0,
    max_shapelet_size=1,
    bootstrap=False,
    metric="euclidean",
    metric_params={"r": 0.2},
)

b.fit(x_train, y_train)
print(b.score(x_test, y_test))
n_node_samples = b.estimators_[1].tree_.n_node_samples
left = b.estimators_[1].tree_.left
right = b.estimators_[1].tree_.right
value = b.estimators_[1].tree_.value
from sklearn.neighbors import KNeighborsRegressor

print(KNeighborsRegressor().fit(x_train, y_train).score(x_test, y_test))



import numpy as np
import matplotlib.pylab as plt

from wildboar.ensemble import ShapeletForestClassifier
from wildboar.datasets import load_dataset
from wildboar.explain.counterfactual import counterfactuals

x_train, x_test, y_train, y_test = load_dataset(
    "TwoLeadECG", repository="wildboar/ucr", merge_train_test=False
)

clf = ShapeletForestClassifier(metric="euclidean", random_state=1, n_estimators=100)
clf.fit(x_train, y_train)

y_pred = clf.predict(x_test)
class_ = clf.classes_[1]
print("Class: %s" % class_)
print("Pred: %r" % y_pred)
print(" - where %r " % (y_pred != class_).nonzero())
x_test = x_test[y_pred != class_][:10]
y_test = y_test[y_pred != class_][:10]

x_counterfactual, success, scores = counterfactuals(
    clf, x_test, class_, random_state=123, scoring="euclidean"
)

print(clf.predict(x_counterfactual))
print(y_test)
print(np.sum(success) / success.shape[0])
print(
    np.sum(clf.predict(x_counterfactual[success]) == y_test[success]) / np.sum(success)
)
fig, ax = plt.subplots(nrows=2)
ax[0].plot(x_counterfactual[0], c="red")
ax[0].plot(x_test[0], c="blue")
ax[1].plot(x_counterfactual[1], c="red")
ax[1].plot(x_test[1], c="blue")
ax[1].legend(["x'", "x"])
plt.show()



import matplotlib.pyplot as plt

from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import cross_validate
from sklearn.pipeline import make_pipeline

from wildboar.datasets import load_dataset
from wildboar.ensemble import ShapeletForestEmbedding

random_state = 1234
x, y = load_dataset("GunPoint")

pipe = make_pipeline(
    ShapeletForestEmbedding(
        n_shapelets=1,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="scaled_euclidean",
        sparse_output=True,
        max_depth=5,
        random_state=random_state,
        n_jobs=-1,
    ),
    LogisticRegression(solver="newton-cg", random_state=random_state),
)
cv = cross_validate(pipe, x, y, cv=5, scoring="accuracy", n_jobs=1)
plt.plot(cv["test_score"], linestyle="", marker="+")
plt.xlabel("CV fold")
plt.ylabel("CV accuracy")
plt.savefig("../fig/sfe_lr.png")



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import math

import numpy as np
from numpy.testing import assert_allclose, assert_, assert_equal
from sklearn.utils import check_random_state

from wildboar.datasets.outlier import MinorityLabeler, EmmottLabeler


def _new_labeler_data(n_samples, n_timestep, classes, n_classes):
    random_state = check_random_state(123)
    x = random_state.randn(n_samples, n_timestep)
    y = np.repeat(classes, n_classes)
    return x, y


def _run_labeler(labeler, x, y):
    new_x, new_y = labeler.fit_transform(x, y)
    new_x_1, new_y_1 = labeler.transform(x, y)
    return new_x, new_x_1, new_y, new_y_1


class SetupLabeler:
    def test_binary(self):
        x, y = _new_labeler_data(20, 10, [1, 2], [15, 5])
        self._run_labeler_suite(x, y, *(np.unique(y, return_counts=True)))

    def test_multiclass(self):
        x, y = _new_labeler_data(20, 10, [1, 2, 3], [8, 7, 5])
        self._run_labeler_suite(x, y, *(np.unique(y, return_counts=True)))

    def _run_labeler_suite(self, x, y, classes, label_count):
        self._run_labeler_n_outliers(x, y, classes, label_count, 0.05)
        self._run_labeler_n_outliers(x, y, classes, label_count, 0.1)

    def _run_labeler_n_outliers(self, x, y, classes, label_count, n_outliers):
        new_x, new_x_1, new_y, new_y_1 = _run_labeler(
            self._new_labeler(n_outliers), x, y
        )
        assert_allclose(new_x, new_x_1)
        assert_allclose(new_y, new_y_1)

        self._assert_labeler(new_x, new_y, classes, label_count, n_outliers)

    def _new_labeler(self, n_outliers):
        raise NotImplemented()

    def _assert_labeler(self, new_x, new_y, classes, label_count, n_outliers):
        raise NotImplemented()


class TestMinorityLabeler(SetupLabeler):
    def _new_labeler(self, n_outliers):
        return MinorityLabeler(n_outliers=n_outliers, random_state=1)

    def _assert_labeler(self, x, y, classes, label_count, n_outliers):
        minority_label_index = np.argmin(label_count)
        minority_label = classes[minority_label_index]
        majority_label_count = np.sum(label_count[classes != minority_label])
        actual_n_outliers = math.ceil(majority_label_count * n_outliers)

        n_samples = majority_label_count + actual_n_outliers
        y_outlier = -np.ones(actual_n_outliers)
        y_inlier = np.ones(majority_label_count)

        assert_equal(x.shape, [n_samples, x.shape[1]])
        assert_equal(y.shape, [n_samples])
        assert_allclose(y, np.concatenate([y_outlier, y_inlier]))


class TestEmmottLabeler(SetupLabeler):
    def _new_labeler(self, n_outliers):
        return EmmottLabeler(n_outliers=n_outliers, difficulty="any", random_state=1)

    def _assert_labeler(self, x, y, classes, label_count, n_outliers):
        pass



# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

cdef struct DtwExtra:
    double *lower
    double *upper

cdef struct Deque:
    Py_ssize_t *queue
    int size
    int capacity
    int front
    int back

cdef void deque_init(Deque *c, Py_ssize_t capacity) nogil

cdef void deque_destroy(Deque *c) nogil

cdef void deque_push_back(Deque *c, Py_ssize_t v) nogil

cdef void deque_pop_front(Deque *c) nogil

cdef void deque_pop_back(Deque *c) nogil

cdef Py_ssize_t deque_front(Deque *c) nogil

cdef Py_ssize_t deque_back(Deque *c) nogil

cdef Py_ssize_t deque_size(Deque *c) nogil

cdef bint deque_empty(Deque *c) nogil

cdef void _dtw_align(double[:] A, double[:] B, Py_ssize_t warp_width, double[:,:] warp) nogil

cdef double _dtw(Py_ssize_t x_offset, Py_ssize_t x_stride, Py_ssize_t x_length, double *X, double x_mean, double x_std,
                 Py_ssize_t y_offset, Py_ssize_t y_stride, Py_ssize_t y_length, double *Y, double y_mean, double y_std,
                 Py_ssize_t r, double *cost, double *cost_prev) nogil


# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten
import re
import operator

import numpy as np

__all__ = [
    "make_dict_filter",
    "make_str_filter",
    "make_list_filter",
    "make_filter",
]


def __n_dims_filter(verb, dataset, x, y):
    if x.ndims < 3:
        return True
    else:
        return __verb_compare(verb, x.shape[1])


def __casting_op(op, l, r):
    lhs_class = l.__class__
    return op(l, lhs_class(r))


def __verb_compare(verb, value):
    match = re.match(__VERB_PATTERN, verb)
    if match:
        return _OPERATORS[match.group(1)](value, match.group(2))
    else:
        raise ValueError("invalid verb (%s)" % verb)


def __new_composite_filter(filters):
    def f(dataset, x, y):
        for filter in filters:
            if not filter(dataset, x, y):
                return False
        return True

    return f


def make_filter(filter):
    """Create a new filter

    Parameters
    ----------
    filter : str, list or dict
        The filter

    Returns
    -------
    function
        The filter function
    """
    if isinstance(filter, list):
        return make_list_filter(filter)
    elif isinstance(filter, dict):
        return make_dict_filter(dict)
    elif isinstance(filter, str):
        return make_str_filter(filter)
    else:
        raise ValueError("invalid filter (%r)" % filter)


def make_list_filter(filter):
    """Make a new filter based on a list of filter strings

    Parameters
    ----------
    filter : list
        A list of filter strings

    Returns
    -------
    function
        The filter function
    """
    filters = []
    for str_filter in filter:
        filters.append(make_str_filter(str_filter))

    return __new_composite_filter(filters)


def make_dict_filter(filter):
    """Make a new filter

    Parameters
    ----------
    filter : dict
        The dict of [subject] -> [op][verb]

    Returns
    -------
    function
        The filter function
    """
    filters = []
    for subject, verb in filter.items():
        if subject not in _SUBJECTS:
            raise ValueError("invalid subject (%s)" % subject)
        filters.append(lambda dataset, x, y: _SUBJECTS[subject](verb, dataset, x, y))

    return __new_composite_filter(filters)


def make_str_filter(filter):
    """Make a new filter

    Parameters
    ----------
    filter : str
        The filter string [subject][op][verb]

    Returns
    -------
    function
        The filter function
    """
    match = re.match(__SUBJECT_VERB_PATTERN, filter)
    if match:
        subject = match.group(1)
        verb = match.group(2)
        if subject not in _SUBJECTS:
            raise ValueError("invalid subject (%s)" % subject)
    else:
        raise ValueError("invalid filter (%s)" % filter)

    def f(dataset, x, y):
        return _SUBJECTS[subject](verb, dataset, x, y)

    return f


__VERB_PATTERN = re.compile("^(<|<=|>=|>|=|=~)\s*(\w+)$")
__SUBJECT_VERB_PATTERN = re.compile("^(\w+)\s*((?:<|<=|>=|>|=|=~)\s*\w+)$")


_SUBJECTS = {
    "dataset": lambda verb, dataset, x, y: __verb_compare(verb, dataset),
    "n_samples": lambda verb, dataset, x, y: __verb_compare(verb, x.shape[0]),
    "n_timestep": lambda verb, dataset, x, y: __verb_compare(verb, x.shape[-1]),
    "n_dims": __n_dims_filter,
    "n_labels": lambda verb, dataset, x, y: __verb_compare(verb, np.unique(y).shape[0]),
}

_OPERATORS = {
    "=~": lambda x, y: y in x,
    "=": lambda x, y: __casting_op(operator.eq, x, y),
    "<=": lambda x, y: __casting_op(operator.le, x, y),
    ">=": lambda x, y: __casting_op(operator.ge, x, y),
    "<": lambda x, y: __casting_op(operator.lt, x, y),
    ">": lambda x, y: __casting_op(operator.gt, x, y),
}



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

# This implementation is heavily inspired by the UCRSuite.
#
# References
#
#  - Rakthanmanon, et al., Searching and Mining Trillions of Time
#    Series Subsequences under Dynamic Time Warping (2012)
#  - http://www.cs.ucr.edu/~eamonn/UCRsuite.html
cimport numpy as np
import numpy as np

from libc.stdlib cimport malloc
from libc.stdlib cimport free

from libc.math cimport INFINITY
from libc.math cimport sqrt
from libc.math cimport floor

from ._distance cimport TSDatabase

from ._distance cimport TSView
from ._distance cimport ScaledDistanceMeasure
from ._distance cimport DistanceMeasure
from ._distance cimport TSCopy
from .._utils cimport fast_mean_std

from .._utils import check_array_fast


cdef void deque_init(Deque *c, Py_ssize_t capacity) nogil:
    c[0].capacity = capacity
    c[0].size = 0
    c[0].queue = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * capacity)
    c[0].front = 0
    c[0].back = capacity - 1


cdef void deque_reset(Deque *c) nogil:
    c[0].size = 0
    c[0].front = 0
    c[0].back = c[0].capacity - 1


cdef void deque_destroy(Deque *c) nogil:
    free(c[0].queue)


cdef void deque_push_back(Deque *c, Py_ssize_t v) nogil:
    c[0].queue[c[0].back] = v
    c[0].back -= 1
    if c[0].back < 0:
        c[0].back = c[0].capacity - 1

    c[0].size += 1


cdef void deque_pop_front(Deque *c) nogil:
    c[0].front -= 1
    if c[0].front < 0:
        c[0].front = c[0].capacity - 1
    c[0].size -= 1


cdef void deque_pop_back(Deque *c) nogil:
    c[0].back = (c[0].back + 1) % c[0].capacity
    c[0].size -= 1


cdef Py_ssize_t deque_front(Deque *c) nogil:
    cdef int tmp = c[0].front - 1
    if tmp < 0:
        tmp = c[0].capacity - 1
    return c[0].queue[tmp]


cdef Py_ssize_t deque_back(Deque *c) nogil:
    cdef int tmp = (c[0].back + 1) % c[0].capacity
    return c[0].queue[tmp]


cdef bint deque_empty(Deque *c) nogil:
    return c[0].size == 0


cdef Py_ssize_t deque_size(Deque *c) nogil:
    return c[0].size


cdef void find_min_max(
    Py_ssize_t offset,
    Py_ssize_t stride,
    Py_ssize_t length,
    double *T,
    Py_ssize_t r,
    double *lower,
    double *upper,
    Deque *dl,
    Deque *du,
) nogil:
    cdef Py_ssize_t i
    cdef Py_ssize_t k

    cdef double current, prev

    deque_reset(du)
    deque_reset(dl)

    deque_push_back(du, 0)
    deque_push_back(dl, 0)

    for i in range(1, length):
        if i > r:
            k = i - r - 1
            upper[k] = T[offset + stride * deque_front(du)]
            lower[k] = T[offset + stride * deque_front(dl)]

        current = T[offset + stride * i]
        prev = T[offset + stride * (i - 1)]
        if current > prev:
            deque_pop_back(du)
            while (not deque_empty(du) and
                   current > T[offset + stride * deque_back(du)]):
                deque_pop_back(du)
        else:
            deque_pop_back(dl)
            while (not deque_empty(dl) and
                   current < T[offset + stride * deque_back(dl)]):
                deque_pop_back(dl)

        deque_push_back(du, i)
        deque_push_back(dl, i)

        if i == 2 * r + 1 + deque_front(du):
            deque_pop_front(du)
        elif i == 2 * r + 1 + deque_front(dl):
            deque_pop_front(dl)

    for i in range(length, length + r + 1):
        upper[i - r - 1] = T[offset + stride * deque_front(du)]
        lower[i - r - 1] = T[offset + stride * deque_front(dl)]

        if i - deque_front(du) >= 2 * r + 1:
            deque_pop_front(du)
        if i - deque_front(dl) >= 2 * r + 1:
            deque_pop_front(dl)


cdef inline double dist(double x, double y) nogil:
    cdef double s = x - y
    return s * s


cdef double constant_lower_bound(
    Py_ssize_t s_offset,
    Py_ssize_t s_stride,
    double *S,
    double s_mean,
    double s_std,
    Py_ssize_t t_offset,
    Py_ssize_t t_stride,
    double *T,
    double t_mean,
    double t_std,
    Py_ssize_t length,
    double best_dist,
) nogil:
    cdef double t_x0, t_y0, s_x0, s_y0
    cdef double t_x1, ty1, s_x1, s_y1
    cdef double t_x2, t_y2, s_x2, s_y2
    cdef double distance, min_dist
    if t_std == 0:
        return 0
    # first and last in T
    t_x0 = (T[t_offset] - t_mean) / t_std
    t_y0 = (T[t_offset + t_stride * (length - 1)] - t_mean) / t_std

    # first and last in S
    s_x0 = (S[s_offset] - s_mean) / s_std
    s_y0 = (S[s_offset + s_stride * (length - 1)] - s_mean) / s_std

    min_dist = dist(t_x0, s_x0) + dist(t_y0, s_y0)
    if min_dist >= best_dist:
        return min_dist

    t_x1 = (T[t_offset + t_stride * 1] - t_mean) / t_std
    s_x1 = (S[s_offset + s_stride * 1] - s_mean) / s_std
    min_dist += min(
        min(dist(t_x1, s_x0), dist(t_x0, s_x1)),
        dist(t_x1, s_x1))

    if min_dist >= best_dist:
        return min_dist

    t_y1 = (T[t_offset + t_stride * (length - 2)] - t_mean) / t_std
    s_y1 = (S[s_offset + s_stride * (length - 2)] - s_mean) / s_std
    min_dist += min(
        min(dist(t_y1, s_y1), dist(t_y0, s_y1)),
        dist(t_y1, s_y1))

    if min_dist >= best_dist:
        return min_dist

    t_x2 = (T[t_offset + t_stride * 2] - t_mean) / t_std
    s_x2 = (S[s_offset + s_stride * 2] - s_mean) / s_std
    min_dist += min(min(dist(t_x0, s_x2),
                        min(dist(t_x1, s_x2),
                            dist(t_x2, s_x2)),
                        dist(t_x2, s_x1)),
                    dist(t_x2, s_x0))

    if min_dist >= best_dist:
        return min_dist

    t_y2 = (T[t_offset + t_stride * (length - 3)] - t_mean) / t_std
    s_y2 = (S[s_offset + s_stride * (length - 3)] - s_mean) / s_std

    min_dist += min(min(dist(t_y0, s_y2),
                        min(dist(t_y1, s_y2),
                            dist(t_y2, s_y2)),
                        dist(t_y2, s_y1)),
                    dist(t_y2, s_y0))

    return min_dist


cdef double cumulative_bound(
    Py_ssize_t offset,
    Py_ssize_t stride,
    Py_ssize_t length,
    double mean,
    double std,
    double *T,
    double lu_mean,
    double lu_std,
    double *lower,
    double *upper,
    double *cb,
    double best_so_far,
) nogil:
    cdef double min_dist = 0
    cdef double x, d, us, ls
    cdef Py_ssize_t i

    for i in range(0, length):
        if min_dist >= best_so_far:
            break

        x = (T[offset + stride * i] - mean) / std
        us = (upper[i] - lu_mean) / lu_std
        ls = (lower[i] - lu_mean) / lu_std
        if x > us:
            d = dist(x, us)
        elif x < ls:
            d = dist(x, ls)
        else:
            d = 0

        min_dist += d
        cb[i] = d
    return min_dist


cdef inline double inner_scaled_dtw(
    Py_ssize_t s_offset,
    Py_ssize_t s_stride,
    int s_length,
    double s_mean,
    double s_std,
    double *S,
    double mean,
    double std,
    Py_ssize_t x_offset,
    double *X_buffer,
    int r,
    double *cb,
    double *cost,
    double *cost_prev,
    double min_dist,
) nogil:
    cdef int i = 0
    cdef int j = 0
    cdef int k = 0

    cdef double x
    cdef double y
    cdef double z
    cdef double v
    cdef double min_cost, distance
    cdef bint std_zero = std == 0
    cdef bint s_std_zero = s_std == 0

    if std_zero and s_std_zero:
        return 0

    cdef double *cost_tmp
    for i in range(0, 2 * r + 1):
        cost[i] = INFINITY
        cost_prev[i] = INFINITY

    for i in range(0, s_length):
        k = max(0, r - i)
        min_cost = INFINITY
        for j in range(max(0, i - r), min(s_length, i + r + 1)):
            if i == 0 and j == 0:
                v = (S[s_offset] - s_mean) / s_std
                if not std_zero:
                    v -= (X_buffer[x_offset] - mean) / std
                cost[k] = v * v
            else:
                if j - 1 < 0 or k - 1 < 0:
                    y = INFINITY
                else:
                    y = cost[k - 1]

                if i - 1 < 0 or k + 1 > 2 * r:
                    x = INFINITY
                else:
                    x = cost_prev[k + 1]

                if i - 1 < 0 or j - 1 < 0:
                    z = INFINITY
                else:
                    z = cost_prev[k]

                v = (S[s_offset + s_stride * i] - s_mean) / s_std
                if not std_zero:
                    v -= (X_buffer[x_offset + j] - mean) / std

                distance = v * v
                cost[k] = min(min(x, y), z) + distance

            if cost[k] < min_cost:
                min_cost = cost[k]

            k += 1

        if i + r < s_length - 1 and min_cost + cb[i + r + 1] >= min_dist:
            return min_cost + cb[i + r + 1]

        cost_tmp = cost
        cost = cost_prev
        cost_prev = cost_tmp
    return cost_prev[k - 1]


cdef double scaled_dtw_distance(
    Py_ssize_t s_offset,
    Py_ssize_t s_stride,
    Py_ssize_t s_length,
    double s_mean,
    double s_std,
    double *S,
    Py_ssize_t t_offset,
    Py_ssize_t t_stride,
    Py_ssize_t t_length,
    double *T,
    Py_ssize_t r,
    double *X_buffer,
    double *cost,
    double *cost_prev,
    double *s_lower,
    double *s_upper,
    double *t_lower,
    double *t_upper,
    double *cb,
    double *cb_1,
    double *cb_2,
    Py_ssize_t *index,
) nogil:
    cdef double current_value = 0
    cdef double mean = 0
    cdef double std = 0
    cdef double dist = 0
    cdef double min_dist = INFINITY

    cdef double lb_kim
    cdef double lb_k
    cdef double lb_k2

    cdef double ex = 0
    cdef double ex2 = 0

    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef Py_ssize_t k
    cdef Py_ssize_t I
    cdef Py_ssize_t buffer_pos

    for i in range(t_length):
        current_value = T[t_offset + t_stride * i]
        ex += current_value
        ex2 += current_value * current_value

        buffer_pos = i % s_length
        X_buffer[buffer_pos] = current_value
        X_buffer[buffer_pos + s_length] = current_value

        if i >= s_length - 1:
            j = (i + 1) % s_length
            I = i - (s_length - 1)
            mean = ex / s_length
            std = sqrt(ex2 / s_length - mean * mean)
            lb_kim = constant_lower_bound(s_offset, s_stride, S,
                                          s_mean, s_std, j, 1, X_buffer,
                                          mean, std, s_length, min_dist)

            if lb_kim < min_dist:
                lb_k = cumulative_bound(j, 1, s_length, mean, std, X_buffer,
                                        s_mean, s_std, s_lower, s_upper,
                                        cb_1, min_dist)
                if lb_k < min_dist:
                    lb_k2 = cumulative_bound(
                        s_offset, s_stride, s_length, s_mean, s_std, S,
                        mean, std, t_lower + I, t_upper + I, cb_2, min_dist)

                    if lb_k2 < min_dist:
                        if lb_k > lb_k2:
                            cb[s_length - 1] = cb_1[s_length - 1]
                            for k in range(s_length - 2, -1, -1):
                                cb[k] = cb[k + 1] + cb_1[k]
                        else:
                            cb[s_length - 1] = cb_2[s_length - 1]
                            for k in range(s_length - 2, -1, -1):
                                cb[k] = cb[k + 1] + cb_2[k]
                        dist = inner_scaled_dtw(
                            s_offset, s_stride, s_length, s_mean,
                            s_std, S, mean, std, j, X_buffer, r, cb,
                            cost, cost_prev, min_dist)

                        if dist < min_dist:
                            if index != NULL:
                                index[0] = (i + 1) - s_length
                            min_dist = dist

            current_value = X_buffer[j]
            ex -= current_value
            ex2 -= current_value * current_value

    return sqrt(min_dist)


cdef inline double inner_dtw(
    Py_ssize_t s_offset,
    Py_ssize_t s_stride,
    int s_length,
    double *S,
    Py_ssize_t t_offset,
    Py_ssize_t t_stride,
    double *T,
    int r,
    double *cost,
    double *cost_prev,
    double min_dist,
) nogil:
    cdef int i = 0
    cdef int j = 0
    cdef int k = 0

    cdef double x
    cdef double y
    cdef double z
    cdef double v
    cdef double min_cost
    cdef double *cost_tmp

    for i in range(0, 2 * r + 1):
        cost[i] = INFINITY
        cost_prev[i] = INFINITY

    for i in range(0, s_length):
        k = max(0, r - i)
        for j in range(max(0, i - r), min(s_length, i + r + 1)):
            if i == 0 and j == 0:
                v = T[t_offset] - S[s_offset]
                cost[k] = v * v
            else:
                if j - 1 < 0 or k - 1 < 0:
                    y = INFINITY
                else:
                    y = cost[k - 1]

                if i - 1 < 0 or k + 1 > 2 * r:
                    x = INFINITY
                else:
                    x = cost_prev[k + 1]

                if i - 1 < 0 or j - 1 < 0:
                    z = INFINITY
                else:
                    z = cost_prev[k]

                v = T[t_offset + t_stride * i] - S[s_offset + s_stride * j]
                cost[k] = min(min(x, y), z) + v * v

            k += 1
        cost_tmp = cost
        cost = cost_prev
        cost_prev = cost_tmp
    return cost_prev[k - 1]


cdef double dtw_distance(
    Py_ssize_t s_offset,
    Py_ssize_t s_stride,
    Py_ssize_t s_length,
    double *S,
    Py_ssize_t t_offset,
    Py_ssize_t t_stride,
    Py_ssize_t t_length,
    double *T,
    Py_ssize_t r,
    double *cost,
    double *cost_prev,
    Py_ssize_t *index,
) nogil:
    cdef double dist = 0
    cdef double min_dist = INFINITY

    cdef Py_ssize_t i
    for i in range(t_length - s_length + 1):
        dist = inner_dtw(s_offset, s_stride, s_length, S,
                         t_offset + t_stride * i, t_stride, T,
                         r, cost, cost_prev, min_dist)
        if dist < min_dist:
            if index != NULL:
                index[0] = i
            min_dist = dist

    return sqrt(min_dist)


cdef double _dtw(
    Py_ssize_t x_offset,
    Py_ssize_t x_stride,
    Py_ssize_t x_length,
    double *X,
    double x_mean,
    double x_std,
    Py_ssize_t y_offset,
    Py_ssize_t y_stride,
    Py_ssize_t y_length,
    double *Y,
    double y_mean,
    double y_std,
    Py_ssize_t r,
    double *cost,
    double *cost_prev,
) nogil:
    """Dynamic time warping distance
    
    Parameters
    ----------
    x_offset : offset of x
    x_stride : stride of x
    x_length : length of x
    X : data of x
    x_mean : mean of array in x (if 0 ignored)
    x_std : std of array in x (or 1)
    y_offset : offset of y
    y_stride : stride of y
    y_length : length of y
    Y : data of y
    y_mean : mean of array in y (if 0 ignored)
    y_std : std of array in y (or 1)
    r : the warp window
    cost : cost matrix (max(x_length, y_length))
    cost_prev : cost matrix (max(x_length, y_length))

    Returns
    -------
    distance : the distance
    """
    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef Py_ssize_t j_start
    cdef Py_ssize_t j_stop
    cdef double x
    cdef double y
    cdef double z
    cdef double v
    cdef bint y_std_zero = y_std == 0.0
    cdef bint x_std_zero = x_std == 0.0
    if x_std_zero and x_std_zero:
        return 0.0

    v = (X[x_offset] - x_mean) / x_std if not x_std_zero else 0.0
    if not y_std_zero:
        v -= (Y[y_offset] - y_mean) / y_std if not y_std_zero else 0.0

    cost_prev[0] = v * v
    for i in range(1, min(y_length, max(0, y_length - x_length) + r)):
        v = (X[x_offset] - x_mean) / x_std if not x_std_zero else 0.0
        if not y_std_zero:
            v -= (Y[y_offset + y_stride * i] - y_mean) / y_std if not y_std_zero else 0.0
        cost_prev[i] = cost_prev[i - 1] + v * v

    if max(0, y_length - x_length) + r < y_length:
        cost_prev[max(0, y_length - x_length) + r] = INFINITY

    for i in range(1, x_length):
        j_start = max(0, i - max(0, x_length - y_length) - r + 1)
        j_stop = min(y_length, i + max(0, y_length - x_length) + r)
        if j_start > 0:
            cost[j_start - 1] = INFINITY
        for j in range(j_start, j_stop):
            x = cost_prev[j]
            if j > 0:
                y = cost[j - 1]
                z = cost_prev[j - 1]
            else:
                y = INFINITY
                z = INFINITY
            v = (X[x_offset + x_stride * i] - x_mean) / x_std if not x_std_zero else 0.0
            if not y_std_zero:
                v -= (Y[y_offset + y_stride * j] - y_mean) / y_std if not y_std_zero else 0.0
            cost[j] = min(min(x, y), z) + v * v

        if j_stop < y_length:
            cost[j_stop] = INFINITY

        cost, cost_prev = cost_prev, cost
    return cost_prev[y_length - 1]


cdef void _dtw_align(double[:] X, double[:] Y, Py_ssize_t r, double[:,:] out) nogil:
    """Compute the warp alignment """
    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef Py_ssize_t j_start, j_stop
    cdef double v
    cdef double x, y, z
    v = X[0] - Y[0]
    out[0, 0] = v * v
    for i in range(1, min(X.shape[0], r + 1)):
        v = X[i] - Y[0]
        out[i, 0] = out[i - 1, 0] + v * v

    for i in range(1, min(Y.shape[0], max(0, Y.shape[0] - X.shape[0]) + r)):
        v = X[0] - Y[i]
        out[0, i] = out[0, i - 1] + v * v

    if max(0, Y.shape[0] - X.shape[0]) + r < Y.shape[0]:
        out[0, max(0, Y.shape[0] - X.shape[0]) + r] = INFINITY

    for i in range(1, X.shape[0]):
        j_start = max(0, i - max(0, X.shape[0] - Y.shape[0]) - r + 1)
        j_stop = min(Y.shape[0], i + max(0, Y.shape[0] - X.shape[0]) + r)
        if j_start > 0:
            out[i, j_start - 1] = INFINITY
        for j in range(j_start, j_stop):
            v = X[i] - Y[j]
            x = out[i - 1, j]
            y = out[i, j - 1] if j > 0 else INFINITY
            z = out[i - 1, j - 1] if j > 0 else INFINITY
            out[i, j] = min(min(x, y), z) + v * v

        if j_stop < Y.shape[0]:
            out[i, j_stop] = INFINITY


def _dtw_alignment(np.ndarray x, np.ndarray y, Py_ssize_t r, np.ndarray out=None):
    """Computes that dtw alignment matrix

    Parameters
    ----------
    x : ndarray of shape (x_size, )
        The first array

    y : ndarray of shape (y_size, )
        The second array

    r : int
        The warping window size

    out : ndarray of shape (x_size, y_size), optional
        To avoid allocating a new array out can be reused

    Returns
    -------
    alignment : ndarray of shape (x_size, y_size)
        - if `out` is given, alignment is out

        Contains the dtw alignment. Values outside the warping
        window size is undefined.
    """
    if not 0 < r < max(x.shape[0], y.shape[0]):
        raise ValueError("invalid r")
    x = check_array_fast(x, c_order=False)
    y = check_array_fast(y, c_order=False)
    cdef Py_ssize_t x_size = x.shape[0]
    cdef Py_ssize_t y_size = y.shape[0]
    if out is None:
        out = np.empty((x_size, y_size))

    if out.shape[0] < x_size or out.shape[1] < y_size:
        raise ValueError("out has wrong shape, got [%d, %d]" (x_size, y_size))

    _dtw_align(x, y, r, out)
    return out


def _dtw_distance(np.ndarray x, np.ndarray y, Py_ssize_t r, bint scale=False):
    """Compute the DTW of x and y

    Parameters
    ----------
    x : ndarray of shape (x_size)
        First time series

    y : ndarray of shape (y_size)
        Second time series

    r : int
        Size of warping window

    scale : bool, optional
        Standardize the arrays

    Returns
    -------
    dtw_distance, float
        The dynamic time warping distance

    Notes
    -----
    This implementation uses `2 * max(y_size, x_size)` memory.
    """
    if not 0 < r < max(x.shape[0], y.shape[0]):
        raise ValueError("invalid r")
    x = check_array_fast(x)
    y = check_array_fast(y)
    cdef Py_ssize_t x_length = <Py_ssize_t> x.shape[0]
    cdef Py_ssize_t x_stride = <Py_ssize_t> x.strides[0] / <Py_ssize_t> x.itemsize
    cdef double *x_data = <double*> x.data
    cdef Py_ssize_t y_length = <Py_ssize_t> y.shape[0]
    cdef Py_ssize_t y_stride = <Py_ssize_t> y.strides[0] / <Py_ssize_t> y.itemsize
    cdef double *y_data = <double*> y.data
    cdef double x_mean = 0
    cdef double x_std = 1
    cdef double y_mean = 0
    cdef double y_std = 1
    if scale:
        x_mean = np.mean(x)
        x_std = np.std(x)
        y_mean = np.mean(y)
        y_std = np.std(y)
    cdef double *cost = <double*> malloc(sizeof(double) * max(x_length, y_length))
    cdef double *cost_prev = <double*> malloc(sizeof(double) * max(x_length, y_length))
    cdef dist = _dtw(0, x_stride, x_length, x_data, x_mean, x_std,
                     0, y_stride, y_length, y_data, y_mean, y_std,
                     r, cost, cost_prev)
    free(cost)
    free(cost_prev)
    return sqrt(dist)


def _dtw_pairwise_distance(np.ndarray x, Py_ssize_t r):
    """Compute the distance between pairs of rows in x

    Parameters
    ----------
    x : ndarray of shape (n_samples, n_timestep)
        The time series

    r : int
        The warping window

    Returns
    -------
    pairwise_distance : ndarray of shape (n_samples, n_samples)
        The distances between pairs of samples (i, j)
    """
    if not 0 < r < x.shape[1]:
        raise ValueError("invalid r")
    x = check_array_fast(x, ensure_2d=True)
    cdef Py_ssize_t length = x.shape[1]
    cdef np.ndarray dists = np.empty((x.shape[0], x.shape[0]), dtype=np.float64)
    cdef Py_ssize_t i, j
    cdef Py_ssize_t i_offset, j_offset
    cdef Py_ssize_t sample_stride = <Py_ssize_t> x.strides[0] / <Py_ssize_t> x.itemsize
    cdef Py_ssize_t timestep_stride = <Py_ssize_t> x.strides[1] / <Py_ssize_t> x.itemsize
    cdef double *data = <double*> x.data
    cdef double *cost = <double*> malloc(sizeof(double) * length)
    cdef double *cost_prev = <double*> malloc(sizeof(double) * length)
    cdef double dist
    for i in range(x.shape[0]):
        i_offset = i * sample_stride
        dists[i, i] = 0
        for j in range(i + 1, x.shape[0]):
            j_offset = j * sample_stride
            dist = sqrt(_dtw(i_offset, timestep_stride, length, data, 0, 1,
                             j_offset, timestep_stride, length, data, 0, 1,
                             r, cost, cost_prev))
            dists[i, j] = dist
            dists[j, i] = dist

    free(cost)
    free(cost_prev)
    return dists


def _dtw_envelop(np.ndarray x, Py_ssize_t r):
    if not 0 < r < x.shape[0]:
        raise ValueError("invalid r")
    x = check_array_fast(x)

    cdef Deque du
    cdef Deque dl
    cdef Py_ssize_t length = x.shape[0]
    cdef Py_ssize_t stride = x.strides[0] / <Py_ssize_t> x.itemsize
    cdef double *data = <double*> x.data
    cdef np.ndarray lower = np.empty(length, dtype=np.float64)
    cdef np.ndarray upper = np.empty(length, dtype=np.float64)
    cdef double *lower_data = <double*> lower.data
    cdef double *upper_data = <double*> upper.data

    deque_init(&dl, 2 * r + 2)
    deque_init(&du, 2 * r + 2)
    find_min_max(0, stride, length, data, r, lower_data, upper_data, &dl, &du)

    deque_destroy(&dl)
    deque_destroy(&du)
    return lower, upper


def _dtw_lb_keogh(np.ndarray x, np.ndarray lower, np.ndarray upper, Py_ssize_t r):
    if not 0 < r < x.shape[0]:
        raise ValueError("invalid r")
    x = check_array_fast(x)
    lower = check_array_fast(lower)
    upper = check_array_fast(upper)
    cdef Py_ssize_t i
    cdef Py_ssize_t length = x.shape[0]
    cdef Py_ssize_t stride = x.strides[0] / <Py_ssize_t> x.itemsize
    cdef double *data = <double*> x.data
    cdef double *lower_data
    cdef double *upper_data
    if lower.strides[0] / <Py_ssize_t> lower.itemsize == 1:
        lower_data = <double*> lower.data
    else:
        lower_data = <double*> malloc(sizeof(double) * length)
        for i in range(length):
            lower_data[i] = lower[i]

    if upper.strides[0] / <Py_ssize_t> lower.itemsize == 1:
        upper_data = <double*> upper.data
    else:
        upper_data = <double*> malloc(sizeof(double) * length)
        for i in range(length):
            upper_data[i] = upper[i]
    cdef np.ndarray cb = np.empty(length, dtype=np.float64)
    cdef double *cb_data = <double*> cb.data
    cdef double min_dist
    min_dist = cumulative_bound(0, stride, length, 0, 1, data, 0, 1, lower_data, upper_data, cb_data, INFINITY)
    if <double*> upper.data != upper_data:
        free(upper_data)
    if <double*> lower.data != lower_data:
        free(lower_data)
    return sqrt(min_dist), cb


cdef inline Py_ssize_t _compute_warp_width(Py_ssize_t length, double r) nogil:
    # Warping path should be [0, length - 1]
    if r == 1:
        return length - 1
    if r < 1:
        return <Py_ssize_t> floor(length * r)
    else:
        return <Py_ssize_t> min(floor(r), length - 1)


cdef class ScaledDtwDistance(ScaledDistanceMeasure):
    cdef double *X_buffer
    cdef double *lower
    cdef double *upper
    cdef double *cost
    cdef double *cost_prev
    cdef double *cb
    cdef double *cb_1
    cdef double *cb_2

    cdef Deque du
    cdef Deque dl

    # The default and minimum size of the cost matrices
    cdef Py_ssize_t cost_size
    cdef double r


    def __cinit__(self, Py_ssize_t n_timestep, double r=0):
        super().__init__(n_timestep)
        if r < 0:
            raise ValueError("illegal warp width")
        self.r = r
        self.cost_size = _compute_warp_width(n_timestep, self.r) * 2 + 1
        self.X_buffer = <double*> malloc(sizeof(double) * n_timestep * 2)
        self.lower = <double*> malloc(sizeof(double) * n_timestep)
        self.upper = <double*> malloc(sizeof(double) * n_timestep)
        self.cost = <double*> malloc(sizeof(double) * self.cost_size)
        self.cost_prev = <double*> malloc(sizeof(double) * self.cost_size)
        self.cb = <double*> malloc(sizeof(double) * n_timestep)
        self.cb_1 = <double*> malloc(sizeof(double) * n_timestep)
        self.cb_2 = <double*> malloc(sizeof(double) * n_timestep)

        if (
            self.X_buffer == NULL or
            self.lower == NULL or
            self.upper == NULL or
            self.cost == NULL or
            self.cost_prev == NULL or
            self.cb == NULL or
            self.cb_1 == NULL or
            self.cb_2 == NULL
        ):
            raise MemoryError()

        deque_init(&self.dl, 2 * _compute_warp_width(n_timestep, self.r) + 2)
        deque_init(&self.du, 2 * _compute_warp_width(n_timestep, self.r) + 2)


    def __reduce__(self):
        return self.__class__, (self.n_timestep, self.r)


    def __dealloc__(self):
        free(self.X_buffer)
        free(self.lower)
        free(self.upper)
        free(self.cost)
        free(self.cost_prev)
        free(self.cb)
        free(self.cb_1)
        free(self.cb_2)
        deque_destroy(&self.dl)
        deque_destroy(&self.du)


    cdef int init_ts_view(
        self,
        TSDatabase *td_ptr,
        TSView *ts_view,
        Py_ssize_t index,
        Py_ssize_t start,
        Py_ssize_t length,
        Py_ssize_t dim,
    ) nogil:
        cdef TSDatabase td = td_ptr[0]
        ScaledDistanceMeasure.init_ts_view(
            self, td_ptr, ts_view, index, start, length, dim
        )

        cdef DtwExtra *dtw_extra = <DtwExtra*> malloc(sizeof(DtwExtra))
        dtw_extra[0].lower = <double*> malloc(sizeof(double) * length)
        dtw_extra[0].upper = <double*> malloc(sizeof(double) * length)

        cdef Py_ssize_t shapelet_offset = (
            index * td.sample_stride +
            start * td.timestep_stride +
            dim * td.dim_stride
        )
        cdef Py_ssize_t warp_width = _compute_warp_width(length, self.r)
        find_min_max(
            shapelet_offset,
            td.timestep_stride,
            length,
            td.data,
            warp_width,
            dtw_extra[0].lower,
            dtw_extra[0].upper,
            &self.dl,
            &self.du,
        )

        ts_view[0].extra = dtw_extra
        return 0


    cdef int init_ts_copy(self, TSCopy *shapelet, TSView *tv_ptr, TSDatabase *td_ptr) nogil:
        cdef int err = ScaledDistanceMeasure.init_ts_copy(
            self, shapelet, tv_ptr, td_ptr
        )
        if err == -1:
            return -1

        cdef TSDatabase td = td_ptr[0]
        cdef TSView shapelet_info = tv_ptr[0]

        cdef DtwExtra *dtw_extra = <DtwExtra*> malloc(sizeof(DtwExtra))
        cdef Py_ssize_t length = shapelet[0].length
        dtw_extra[0].lower = <double*> malloc(sizeof(double) * length)
        dtw_extra[0].upper = <double*> malloc(sizeof(double) * length)

        cdef Py_ssize_t warp_width = _compute_warp_width(length, self.r)
        find_min_max(
            0,
            1,
            length,
            shapelet[0].data,
            warp_width,
            dtw_extra[0].lower,
            dtw_extra[0].upper,
            &self.dl,
            &self.du,
        )
        shapelet[0].extra = dtw_extra
        return 0


    cdef int init_ts_copy_from_ndarray(self, TSCopy *tc, np.ndarray arr, Py_ssize_t dim):
        cdef int err = ScaledDistanceMeasure.init_ts_copy_from_ndarray(
            self, tc, arr, dim
        )
        if err == -1:
            return -1
        cdef Py_ssize_t length = tc[0].length
        cdef DtwExtra *dtw_extra = <DtwExtra*> malloc(sizeof(DtwExtra))
        dtw_extra[0].lower = <double*> malloc(sizeof(double) * length)
        dtw_extra[0].upper = <double*> malloc(sizeof(double) * length)

        cdef Py_ssize_t warp_width = _compute_warp_width(length, self.r)
        find_min_max(
            0,
            1,
            length,
            tc[0].data,
            warp_width,
            dtw_extra[0].lower,
            dtw_extra[0].upper,
            &self.dl,
            &self.du,
        )
        tc[0].extra = dtw_extra
        return 0


    cdef double ts_copy_sub_distance(
        self,
        TSCopy *s,
        TSDatabase *td,
        Py_ssize_t t_index,
        Py_ssize_t *return_index=NULL,
    ) nogil:
        cdef Py_ssize_t sample_offset = (t_index * td.sample_stride + s.dim * td.dim_stride)
        cdef double *s_lower
        cdef double *s_upper
        cdef DtwExtra *extra
        cdef Py_ssize_t warp_width = _compute_warp_width(s.length, self.r)

        if s.extra != NULL:
            extra = <DtwExtra*> s.extra
            s_lower = extra[0].lower
            s_upper = extra[0].upper
        else:
            s_lower = <double*> malloc(sizeof(double) * s.length)
            s_upper = <double*> malloc(sizeof(double) * s.length)

            find_min_max(
                0,
                1,
                s.length,
                s.data,
                warp_width,
                s_lower,
                s_upper,
                &self.dl,
                &self.du,
            )

        find_min_max(
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            warp_width,
            self.lower,
            self.upper,
            &self.dl,
            &self.du,
        )

        cdef double distance = scaled_dtw_distance(
            0,
            1,
            s.length,
            s.mean,
            s.std,
            s.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            warp_width,
            self.X_buffer,
            self.cost,
            self.cost_prev,
            s_lower,
            s_upper,
            self.lower,
            self.upper,
            self.cb,
            self.cb_1,
            self.cb_2,
            return_index,
        )
        if s.extra == NULL:
            free(s_lower)
            free(s_upper)

        return distance


    cdef double ts_view_sub_distance(self, TSView *s, TSDatabase *td, Py_ssize_t t_index) nogil:
        cdef Py_ssize_t sample_offset = (t_index * td.sample_stride + s.dim * td.dim_stride)
        cdef Py_ssize_t shapelet_offset = (
            s.index * td.sample_stride +
            s.dim * td.dim_stride +
            s.start * td.timestep_stride
        )

        cdef Py_ssize_t warp_width = _compute_warp_width(s.length, self.r)
        cdef DtwExtra *dtw_extra = <DtwExtra*> s.extra
        find_min_max(
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            warp_width,
            self.lower,
            self.upper,
            &self.dl,
            &self.du,
        )

        return scaled_dtw_distance(
            shapelet_offset,
            td.timestep_stride,
            s.length,
            s.mean,
            s.std,
            td.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            warp_width,
            self.X_buffer,
            self.cost,
            self.cost_prev,
            dtw_extra[0].lower,
            dtw_extra[0].upper,
            self.lower,
            self.upper,
            self.cb,
            self.cb_1,
            self.cb_2,
            NULL,
        )


    cdef double ts_copy_distance(self, TSCopy *s, TSDatabase *td, Py_ssize_t t_index) nogil:
        cdef Py_ssize_t sample_offset = (t_index * td.sample_stride + s.dim * td.dim_stride)
        cdef Py_ssize_t warp_width = max(_compute_warp_width(s.length, self.r), 1)

        cdef Py_ssize_t max_length = max(s.length, td.n_timestep)
        if max_length > self.cost_size:
            free(self.cost)
            free(self.cost_prev)
            self.cost = <double*> malloc(sizeof(double) * max_length)
            self.cost_prev = <double*> malloc(sizeof(double) * max_length)
            if self.cost == NULL or self.cost_prev == NULL:
                with gil:
                    raise MemoryError()

        cdef double t_mean, t_std
        fast_mean_std(
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            &t_mean,
            &t_std,
        )

        cdef double dist = _dtw(
            0,
            1,
            s.length,
            s.data,
            s.mean, # assuming TSCopy is initialized with self
            s.std,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            t_mean,
            t_std,
            warp_width,
            self.cost,
            self.cost_prev
        )
        return sqrt(dist)

    cdef bint support_unaligned(self) nogil:
        return True


cdef class DtwDistance(DistanceMeasure):
    cdef double *cost
    cdef double *cost_prev
    cdef double r
    cdef Py_ssize_t cost_size


    def __cinit__(self, Py_ssize_t n_timestep, double r=0):
        super().__init__(n_timestep)
        if r < 0:
            raise ValueError("illegal warp width")
        self.r = r
        self.cost_size = _compute_warp_width(n_timestep, self.r) * 2 + 1
        self.cost = <double*> malloc(sizeof(double) * self.cost_size)
        self.cost_prev = <double*> malloc(sizeof(double) * self.cost_size)

        if self.cost == NULL or self.cost_prev == NULL:
            raise MemoryError()


    def __dealloc__(self):
        free(self.cost)
        free(self.cost_prev)


    def __reduce__(self):
        return self.__class__, (self.n_timestep, self.r)


    cdef double ts_copy_sub_distance(
        self,
        TSCopy *s,
        TSDatabase *td,
        Py_ssize_t t_index,
        Py_ssize_t *return_index = NULL,
    ) nogil:
        cdef Py_ssize_t sample_offset = (t_index * td.sample_stride + s.dim * td.dim_stride)
        cdef Py_ssize_t warp_width = _compute_warp_width(s.length, self.r)

        return dtw_distance(
            0,
            1,
            s.length,
            s.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            warp_width,
            self.cost,
            self.cost_prev,
            return_index,
        )


    cdef double ts_view_sub_distance(
            self, TSView *ts_ptr, TSDatabase *td_ptr, Py_ssize_t t_index) nogil:
        cdef TSDatabase td = td_ptr[0]
        cdef TSView s = ts_ptr[0]
        cdef Py_ssize_t sample_offset = (t_index * td.sample_stride +
                                     s.dim * td.dim_stride)
        cdef Py_ssize_t shapelet_offset = (s.index * td.sample_stride +
                                       s.dim * td.dim_stride +
                                       s.start * td.timestep_stride)
        cdef Py_ssize_t warp_width = _compute_warp_width(s.length, self.r)
        return dtw_distance(
            shapelet_offset,
            td.timestep_stride,
            s.length,
            td.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            warp_width,
            self.cost,
            self.cost_prev,
            NULL)


    cdef int ts_copy_matches(
        self,
        TSCopy *s,
        TSDatabase *td,
        Py_ssize_t t_index,
        double threshold,
        Py_ssize_t** matches,
        double** distances,
        Py_ssize_tn_matches,
    ) nogil except -1:
        with gil:
            raise NotImplemented()

    cdef double ts_copy_distance(self, TSCopy *s, TSDatabase *td, Py_ssize_t t_index) nogil:
        cdef Py_ssize_t sample_offset = (t_index * td.sample_stride + s.dim * td.dim_stride)
        cdef Py_ssize_t warp_width = max(_compute_warp_width(s.length, self.r), 1)
        cdef Py_ssize_t max_length = max(s.length, td.n_timestep)

        if max_length > self.cost_size:
            free(self.cost)
            free(self.cost_prev)
            self.cost = <double*> malloc(sizeof(double) * max_length)
            self.cost_prev = <double*> malloc(sizeof(double) * max_length)
            if self.cost == NULL or self.cost_prev == NULL:
                with gil: raise MemoryError()

        cdef double dist = _dtw(
            0,
            1,
            s.length,
            s.data,
            0,
            1,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            0,
            1,
            warp_width,
            self.cost,
            self.cost_prev,
        )
        return sqrt(dist)


    cdef bint support_unaligned(self) nogil:
        return True
